from __future__ import annotations

import contextlib
import dataclasses
import functools
import importlib
import os
import pathlib
import re
import types

from typing import TYPE_CHECKING

from mknodes.data import commitconventions, installmethods, taskrunners, tools
from mknodes.info import (
    contexts,
    # githubinfo,
    grifferegistry,
    license,
    mkdocsconfigfile,
    packageregistry,
    pyproject,
    reporegistry,
)
from mknodes.utils import (
    icons,
    log,
    packagehelpers,
    pathhelpers,
    reprhelpers,
    yamlhelpers,
)


if TYPE_CHECKING:
    import griffe

    from griffe.dataclasses import Alias

    from mknodes.info import packageinfo

logger = log.get_logger(__name__)


GITHUB_REGEX = re.compile(
    r"(?:http?:\/\/|https?:\/\/)?"
    r"(?:www\.)?"
    r"github\.com\/"
    r"(?:\/*)"
    r"([\w\-\.]*)\/"
    r"([\w\-]*)"
    r"(?:\/|$)?"  # noqa: COM812
)


@dataclasses.dataclass
class PackageExtra:
    """A class describing a package extra, used to define additional dependencies."""

    name: str
    """Name of the extra."""
    packages: list[str] = dataclasses.field(default_factory=list)
    """List of packages which are part of the extra."""
    description: str = ""
    """Optional description for the extra.

    Must be defined in pyproject mknodes section.
    """


class FolderInfo:
    """Aggregates information about a working dir."""

    def __init__(self, path: str | os.PathLike | None = None):
        """Constructor.

        Arguments:
            path: Path to the repo.
        """
        # packagehelpers.install_or_import(mod_name)
        self.git = reporegistry.get_repo(path or ".")
        self.path = pathlib.Path(self.git.working_dir)
        self.pyproject = pyproject.PyProject(self.path)
        self.mkdocs_config = mkdocsconfigfile.MkDocsConfigFile()
        if (mk_path := self.path / "mkdocs.yml").exists():
            with contextlib.suppress(yamlhelpers.YAMLError):
                self.mkdocs_config = mkdocsconfigfile.MkDocsConfigFile(mk_path)
        # self.github = githubinfo.GitHubRepo(
        #     self.repository_username,
        #     self.repository_name,
        # )
        self._temp_directory = None

    def __fspath__(self):
        return str(self.path)

    @functools.cached_property
    def module(self) -> types.ModuleType:
        """Return the module itself."""
        mod_name = packagehelpers.distribution_to_package(self.git.repo_name)
        return importlib.import_module(mod_name)

    @functools.cached_property
    def griffe_module(self) -> griffe.Module | Alias:
        """Return a griffe Module containing information about the module."""
        # Long-term ideally we would pull all information from here.
        mod_name = packagehelpers.distribution_to_package(self.git.repo_name)
        return grifferegistry.get_module(mod_name)

    def __repr__(self):
        return reprhelpers.get_repr(self, path=self.path)

    @functools.cached_property
    def info(self) -> packageinfo.PackageInfo:
        """Return a PackageInfo object for given distribution."""
        return packageregistry.get_info(self.pyproject.name or self.git.repo_name)

    @functools.cached_property
    def extras(self) -> dict[str, PackageExtra]:
        """Return a dict containing extras and the packages {extra: [package_1, ...]}."""
        dct = {}
        for k, v in self.info.extras.items():
            dct[k] = PackageExtra(
                k,
                packages=v,
                description=self.pyproject.extras_descriptions.get(k, ""),
            )
        return dct

    @functools.cached_property
    def repository_url(self) -> str:
        """Return url of the repository by querying multiple sources.

        Checks MkDocs config file, Git repository info and project metadata.
        """
        if url := self.mkdocs_config.get("repo_url"):
            return url
        if url := self.git.repo_url:
            return url
        if url := self.info.repository_url:
            return url
        msg = "Could not find any repository url"
        raise RuntimeError(msg)

    @functools.cached_property
    def repository_username(self) -> str:
        """Return the username for the remote repository."""
        if match := GITHUB_REGEX.match(self.repository_url or ""):
            return match.group(1)
        msg = "Could not detect repository username"
        raise RuntimeError(msg)

    @functools.cached_property
    def repository_name(self) -> str:
        """Return the name of the remote repository."""
        if match := GITHUB_REGEX.match(self.repository_url or ""):
            return match.group(2)
        msg = "Could not detect repository name"
        raise RuntimeError(msg)

    @functools.cached_property
    def inventory_url(self) -> str | None:
        """Return best guess for a link to an inventory file."""
        if url := self.mkdocs_config.get("site_url"):
            return f"{url.rstrip('/')}/objects.inv"
        return None

    @functools.cached_property
    def package_name(self) -> str:
        """Return name of the package."""
        return self.info.package_name

    @functools.cached_property
    def package_repos(self) -> list[installmethods.InstallMethodStr]:
        """Return a list of package repositories the package is available on."""
        repos = self.pyproject.package_repos or ["pip"]
        if (
            "pip" in repos
            and "pipx" not in repos
            and "console_scripts" in self.info.entry_points
        ):
            repos.append("pipx")
        return repos

    @functools.cached_property
    def commit_types(self) -> list[commitconventions.CommitTypeStr]:
        """Return commit types allowed for code commits."""
        return self.pyproject.allowed_commit_types

    @functools.cached_property
    def tools(self) -> list[tools.Tool]:
        """Return a list of build tools used by this package."""
        return [instance for t in tools.TOOLS.values() if (instance := t(self)).used]

    @functools.cached_property
    def docstring_style(self) -> str | None:
        """Return docstring style (google / numpy)."""
        if style := self.pyproject.docstring_style:
            return style
        if section := self.mkdocs_config.mkdocstrings_config:
            return section.get("options", {}).get("docstring_style")
        return None

    @functools.cached_property
    def license_file_path(self) -> pathlib.Path | None:
        """Return license file path (relative to project root) from metadata."""
        for path in ["LICENSE", "LICENSE.md", "LICENSE.txt"]:
            if (file := self.path / path).exists():
                return file
        if file := self.info.metadata.json.get("license_file"):
            return self.path / file if isinstance(file, str) else self.path / file[0]
        return None

    @functools.cached_property
    def license_text(self) -> str | None:
        """Return full license text.

        Text comes either from a local license file or from a template populated with
        metadata.
        """
        if self.license_file_path:
            return self.license_file_path.read_text(encoding="utf-8")
        if license_name := self.info.license_name:
            lic = license.License.from_name(license_name)
            lic.resolve_by_distribution(self.info.name)
            return lic.content
        return None

    @functools.cached_property
    def social_info(self) -> list[dict[str, str]]:
        result = []
        if url := self.repository_url:
            result.append(
                dict(icon="fa-brands:github", link=url),
            )
        for link in self.info.urls.values():
            if "github" in link or "pypi.org" in link:
                continue
            if icon := icons.icon_for_url(link):
                result.append(dict(icon=icon, link=link))
        result.append(
            dict(
                icon="fa-brands:python",
                link=f"https://pypi.org/project/{self.module.__name__}/",
            ),
        )
        return result

    @functools.cached_property
    def task_runners(self) -> list[taskrunners.TaskRunner]:
        """Return list of task runners used by this package."""
        return [
            runner
            for runner in taskrunners.TASK_RUNNERS.values()
            if any(
                pathhelpers.find_file_in_folder_or_parent(i, self.path)
                for i in runner.filenames
            )
        ]

    @functools.cached_property
    def context(self) -> contexts.PackageContext:
        return contexts.PackageContext(
            pretty_name=self.mkdocs_config.get("site_name") or self.info.name,
            distribution_name=self.info.name,
            version=self.info.version,
            author_name=self.info.author_name,
            author_email=self.info.author_email,
            docstring_style=self.docstring_style,
            description=self.info.metadata["Description"],
            summary=self.info.metadata["Summary"],
            authors=self.info.authors,
            module=self.module,
            griffe_module=self.griffe_module,
            urls=self.info.urls,
            classifiers=self.info.classifiers,
            classifier_map=self.info.classifier_map,
            keywords=self.info.keywords,
            license_name=self.info.license_name,
            license_text=self.license_text,
            required_python_version=self.info.required_python_version,
            required_packages=self.info.required_packages,
            required_package_names=self.info.required_package_names,
            extras=self.extras,
            tools=self.tools,
            entry_points=self.info.entry_points,
            cli=self.info.cli,
            cli_info=self.info.cli_info,
            mkdocs_config=self.mkdocs_config,
            pyproject_file=self.pyproject,
            social_info=self.social_info,
            repository_path=self.path,
            repository_url=self.repository_url,
            repository_username=self.repository_username,
            repository_name=self.repository_name,
            inventory_url=self.inventory_url,
            task_runners=self.task_runners,
            build_system=self.pyproject.build_system,
            configured_build_systems=self.pyproject.configured_build_systems,
            tool_section=self.pyproject.tool,
            commit_types=self.pyproject.allowed_commit_types,
            package_repos=self.pyproject.package_repos,
            line_length=self.pyproject.line_length,
        )


if __name__ == "__main__":
    info = FolderInfo("https://github.com/mkdocs/mkdocs.git")
    print(info.context)
    log.basic()



from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
import contextlib
import functools
import importlib
import importlib.util
import inspect
import os
import pathlib
import sys
import types
import typing

from mknodes.utils import log


T = typing.TypeVar("T", bound=type)


logger = log.get_logger(__name__)


def iter_subclasses(
    klass: T,
    *,
    recursive: bool = True,
    filter_abstract: bool = False,
    filter_generic: bool = True,
    filter_locals: bool = True,
) -> typing.Iterator[T]:
    """Recursively iter all subclasses of given klass.

    Arguments:
        klass: class to get subclasses from
        filter_abstract: whether abstract base classes should be included.
        filter_generic: whether generic base classes should be included.
        filter_locals: whether local base classes should be included.
        recursive: whether to also get subclasses of subclasses.
    """
    if getattr(klass.__subclasses__, "__self__", None) is None:
        return
    for kls in klass.__subclasses__():
        if recursive:
            yield from iter_subclasses(kls)
        if filter_abstract and inspect.isabstract(kls):
            continue
        if filter_generic and kls.__qualname__.endswith("]"):
            continue
        if filter_locals and "<locals>" in kls.__qualname__:
            continue
        yield kls


@typing.overload
def to_module(module, return_none: typing.Literal[False] = ...) -> types.ModuleType: ...


@typing.overload
def to_module(
    module,
    return_none: typing.Literal[True] = ...,
) -> types.ModuleType | None: ...


@functools.cache
def to_module(
    module: str | Sequence[str] | types.ModuleType,
    return_none: bool = True,
) -> types.ModuleType | None:
    """Return a module for given module path. If module is given, just return it.

    Arguments:
        module: A ModuleType, str or sequence.
        return_none: In case module cant get imported, return None.
                     If False, Exception is thrown.
    """
    match module:
        case (str(), *_) | str():  # type: ignore[attr-defined]
            module_path = module if isinstance(module, str) else ".".join(module)
            try:
                return import_module(module_path)
            except (ImportError, AttributeError) as e:
                logger.warning("Could not import %s", module_path)
                if return_none:
                    return None
                raise ImportError from e
        case types.ModuleType():
            return module
        case _:
            raise TypeError(module)


@functools.cache
def to_class(klass: type | str | tuple[str, ...] | list[str]):
    """Convert given input to a class.

    If input is a string or Sequence, interpret it as a dotted path.

    Arguments:
        klass: Name / path to get a klass for
    """
    match klass:
        case type():
            return klass
        case str() if ":" in klass:
            # path.to.mod:Classname
            mod_path, klass_name = klass.split(":", maxsplit=1)
            mod = import_module(mod_path)
            return getattr(mod, klass_name)
        case str():
            # path.to.mod.Classname
            parts = klass.split(".")
            mod = import_module(".".join(parts[:-1]))
            return getattr(mod, parts[-1])
        case tuple() | list():
            mod = import_module(".".join(klass[:-1]))
            return getattr(mod, klass[-1])
        case _:
            raise TypeError(klass)


def to_module_parts(  # type: ignore
    module: Sequence[str] | str | types.ModuleType,
) -> tuple[str, ...]:
    """Returns a tuple describing the module path.

    Result is of form ("module", "submodule", "subsubmodule")

    Arguments:
        module: A ModuleType, str or sequence.
    """
    match module:
        case (str(), *_):  # type: ignore[attr-defined]
            return tuple(module)
        case str():
            return tuple(module.split("."))
        case types.ModuleType():
            return tuple(module.__name__.split("."))
        case pathlib.Path() if not module.is_absolute():
            return module.parts
        case _:
            raise TypeError(module)


def to_dotted_path(
    obj: Sequence[str] | str | types.ModuleType | types.MethodType | type,
) -> str:
    """Return dotted path for given input.

    Arguments:
        obj: Input to return dotted path for
    """
    match obj:
        case (str(), *_):  # type: ignore[attr-defined]
            return ".".join(obj)
        case str():
            return obj
        case types.ModuleType():
            return obj.__name__
        case type() | Callable():
            return f"{obj.__module__}.{obj.__qualname__}"
        case _:
            raise TypeError(obj)


@functools.cache
def iter_classes(
    module: types.ModuleType | str | tuple[str, ...],
    *,
    type_filter: type | None | types.UnionType = None,
    module_filter: str | None = None,
    filter_by___all__: bool = False,
    recursive: bool = False,
) -> Iterator[type]:
    """Yield classes from given module.

    Arguments:
        module: either a module or a path to a module in form of str or
                tuple of strings.
        type_filter: only yield classes which are subclasses of given type.
        module_filter: filter by a module prefix.
        filter_by___all__: Whether to filter based on whats defined in __all__.
        recursive: import all submodules recursively and also yield their classes.
    """
    mod = to_module(module)
    if not mod:
        return []
    if recursive:
        for _name, submod in get_members(mod, inspect.ismodule):
            if submod.__name__.startswith(module_filter or ""):
                yield from iter_classes(
                    submod,
                    type_filter=type_filter,
                    module_filter=submod.__name__,
                    filter_by___all__=filter_by___all__,
                    recursive=True,
                )
    for klass_name, kls in get_members(mod, inspect.isclass):
        has_all = hasattr(mod, "__all__")
        if filter_by___all__ and (not has_all or klass_name not in mod.__all__):
            continue
        if type_filter is not None and not issubclass(kls, type_filter):
            continue
        if module_filter is not None and not kls.__module__.startswith(module_filter):
            continue
        yield kls


@functools.cache
def get_topmost_module_path(obj: Callable) -> str:
    """Return path of topmost module containing given class.

    If a class is imported in any of its parent modules, return that "shorter" path.

    So for a class "submodule.classmodule.Class", it could return "submodule.Class"

    Arguments:
        obj: Klass to get the path for.
    """
    to_search_for = obj.__self__ if hasattr(obj, "__self__") else obj
    path = obj.__module__
    parts = path.split(".")
    while parts:
        with contextlib.suppress(TypeError):
            new_path = ".".join(parts)
            mod = import_module(new_path)
            objs = [i for _i_name, i in get_members(mod)]
            if to_search_for in objs:
                path = new_path
        parts = parts[:-1]
    return path


@functools.cache
def get_submodules(
    module: types.ModuleType | str | tuple[str, ...],
) -> list[types.ModuleType]:
    """Return list of submodules of given module.

    Arguments:
        module: Module to return submodules from.
    """
    module = to_module(module)
    return [
        mod
        for name, mod in get_members(module, inspect.ismodule)
        if name.startswith(module.__name__)
    ]


@functools.cache
def import_module(mod: str) -> types.ModuleType:
    """Cached version of importlib.import_module.

    Arguments:
        mod: The module to import.
    """
    return importlib.import_module(mod)


@functools.cache
def get_members(obj: object, predicate: Callable | None = None):
    """Cached version of inspect.getmembers.

    Arguments:
        obj: Object to get members for
        predicate: Optional predicate for the members
    """
    return inspect.getmembers(obj, predicate)


@functools.cache
def import_file(path: str | os.PathLike) -> types.ModuleType:
    """Import a module based on a file path.

    Arguments:
        path: Path which should get imported
    """
    module_name = pathlib.Path(path).stem
    spec = importlib.util.spec_from_file_location(module_name, path)
    if spec is None:
        raise RuntimeError
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)  # type: ignore[union-attr]
    return module


@functools.cache
def to_callable(path: str | Callable) -> Callable:
    """Return a callable from a string describing the path to a Callable.

    If path already is a callable, return it without changes.
    Must be of format "module.path:object.fn"

    Arguments:
        path: The path to the callable to return.
    """
    if callable(path):
        return path
    modname, _sep, qualname = path.partition(":")
    obj = import_file(modname) if modname.endswith(".py") else import_module(modname)
    for attr in qualname.split("."):
        obj = getattr(obj, attr)
    if not callable(obj):
        msg = "Incorrect path"
        raise TypeError(msg)
    return obj


def get_code_name(obj) -> str:
    """Get a title for an object representing code.

    Arguments:
        obj: The object to get a name for.
    """
    match obj:
        case types.CodeType():
            return obj.co_name
        case types.TracebackType():
            return obj.tb_frame.f_code.co_name
        case types.FrameType():
            return obj.f_code.co_name
        case Callable():
            return to_dotted_path(obj)
        case _:
            return obj.__name__

    # def collect_modules(
    #     self,
    #     *,
    #     recursive: bool = False,
    #     predicate: Callable[[types.ModuleType], bool] | None = None,
    #     submodule: types.ModuleType | str | tuple | list | None = None,
    # ):
    #     """Collect submodules.

    #     Arguments:
    #         recursive: Collect recursively
    #         predicate: Module filter predicate
    #         submodule: Module to collect from. If None, collect from project module.
    #     """
    #     for module in self.iter_modules(
    #         recursive=recursive,
    #         predicate=predicate,
    #         submodule=submodule,
    #     ):
    #         self.submodules.add(module)

    # def iter_modules(
    #     self,
    #     *,
    #     submodule: types.ModuleType | str | tuple | list | None = None,
    #     recursive: bool = False,
    #     predicate: Callable[[types.ModuleType], bool] | None = None,
    #     _seen: set | None = None,
    # ) -> Iterator[types.ModuleType]:
    #     """Iterate over all submodules of the module.

    #     Arguments:
    #         submodule: filter based on a submodule
    #         recursive: whether to only iterate over members of current module
    #                    or whether it should also include modules from submodules.
    #         predicate: filter modules based on a predicate.
    #     """
    #     mod = classhelpers.to_module(submodule) if submodule else self.module
    #     seen = _seen or set()
    #     if mod is None:
    #         return
    #     for submod_name, submod in inspect.getmembers(mod, inspect.ismodule):
    #         not_in_all = hasattr(mod, "__all__") and submod_name not in mod.__all__
    #         filtered_by_all = self.filter_by___all__ and not_in_all
    #         not_filtered_by_pred = predicate(submod) if predicate else True
    #         # if self.module_name in mod.__name__.split(".")
    #         if not filtered_by_all and not_filtered_by_pred:
    #             yield submod
    #         if recursive and submod not in seen:
    #             seen.add(submod)
    #             yield from self.iter_modules(
    #                 submodule=submod,
    #                 recursive=True,
    #                 predicate=predicate,
    #                 _seen=seen,
    #             )


if __name__ == "__main__":
    import mknodes as mk

    print(get_topmost_module_path(mk.MkCode.for_object))
    print(mk.MkCode.for_object.__qualname__)



from __future__ import annotations

import dataclasses
import functools
import json
import pathlib

from typing import Any, Self

from mknodes import paths
from mknodes.info import packageregistry
from mknodes.jinja import environment
from mknodes.utils import log


logger = log.get_logger(__name__)


@functools.cache
def get_db() -> list[dict[str, Any]]:
    """Return dictionary with license data."""
    path = paths.RESOURCES / "licenses" / "db.json"
    with path.open("r") as file:
        return json.load(file)["licenses"]


@dataclasses.dataclass
class License:
    """Class representing a license.

    Arguments:
        name: Name of the license
        identifier: License identifier
        content: License text content
        path: Path to the license
        sources: License sources
        osi_approved: Whether license is OSI-approved
        header: Optional header for the license.
    """

    name: str
    identifier: str
    content: str
    path: str | None = None
    sources: list | None = None
    osi_approved: bool | None = None
    header: str | None = None

    @classmethod
    def from_name(cls, name_or_id: str) -> Self:
        """Get license based on license name.

        Arguments:
            name_or_id: Name or id of the license to get.
        """
        db = get_db()
        name_or_id = name_or_id.lower()
        for lic in db:
            if name_or_id in {lic["name"].lower(), lic["id"].lower()}:
                path = paths.RESOURCES / "licenses" / "templates" / lic["template"]
                return cls(
                    name=lic["name"],
                    identifier=lic["id"],
                    content=path.read_text(encoding="utf-8"),
                    path=path,
                    sources=lic["sources"],
                    osi_approved=lic["osi_approved"],
                    header=lic["header"],
                )
        raise ValueError(name_or_id)

    @classmethod
    def from_path(cls, path: str) -> Self:
        """Get a license from a file path.

        Arguments:
            path: Path to get license from.
        """
        p = pathlib.Path(path)
        return cls(
            path=str(p),
            content=p.read_text(encoding="utf-8"),
            name=p.name,
            identifier=p.name,
        )

    def resolve_by_distribution(self, distribution: str):
        """Resolve license based on distribution data.

        Arguments:
            distribution: Distribution to use data from.
        """
        info = packageregistry.get_info(distribution)
        env = environment.Environment()

        class Ctx:
            copyright_holder = info.author_name
            organization = info.author_name
            program_description = info.metadata["Summary"]
            program_name = info.name
            program_url = info.repository_url or ""
            program_version = info.metadata["Version"]
            email = info.author_email

        env.globals["metadata"] = Ctx
        self.content = env.render_string(self.content)

    def resolve_template(
        self,
        holder: str,
        package_name: str,
        org: str | None = None,
        website: str = "",
        mail_address: str = "",
        summary: str = "",
        version: str = "",
    ):
        """Resolve license template by manually passing needed metadata.

        Arguments:
            holder: Copyright holder
            package_name: Name of the package the license is used for
            org: Name of the organization
            website: Website of the copyright holders
            mail_address: Mail address of the copyright holder
            summary: Summary of the program the license is used for
            version: Version of the program the license is used for
        """
        env = environment.Environment()

        class Ctx:
            copyright_holder = holder
            organization = org or holder
            program_description = summary
            program_name = package_name
            program_url = website
            program_version = version
            email = mail_address

        env.globals["metadata"] = Ctx
        self.content = env.render_string(self.content)


if __name__ == "__main__":
    db = License.from_name("BSD-3-Clause")
    db.resolve_by_distribution("mknodes")
    print(db)



from __future__ import annotations

import os
import pathlib

from mknodes.info import configfile
from mknodes.utils import log, mergehelpers, yamlhelpers


logger = log.get_logger(__name__)


class YamlFile(configfile.ConfigFile):
    def __init__(
        self,
        path: str | os.PathLike | None = None,
        mode: yamlhelpers.LoaderStr = "unsafe",
        resolve_inherit_tag: bool = False,
    ):
        super().__init__(path)
        if resolve_inherit_tag:
            self.resolve_inherit_tag(mode)

    def resolve_inherit_tag(
        self,
        mode: yamlhelpers.LoaderStr = "unsafe",
    ):
        """Resolve INHERIT key-value pair for this YAML file.

        If this YAML file contains a key-value pair like "INHERIT: path_to_config.yml",
        this method will resolve that tag by using the config at given path as the
        "parent config".

        Also supports a list of files for INHERIT.

        Arguments:
            mode: The Yaml loader type
        """
        if not self.path:
            msg = "Config file needs file path (INHERIT path is relative to file path)"
            raise ValueError(msg)
        abspath = pathlib.Path(self.path).absolute()
        if "INHERIT" not in self._data:
            return
        file_path = self._data.pop("INHERIT")
        file_paths = [file_path] if isinstance(file_path, str) else file_path
        for path in file_paths:
            parent_cfg = abspath.parent / path
            logger.debug("Loading inherited configuration file: %s", parent_cfg)
            text = parent_cfg.read_text()
            parent = yamlhelpers.load_yaml(text, mode)
            self._data = mergehelpers.merge_dicts(parent, self._data)

    @classmethod
    def _dump(cls, data: dict) -> str:
        return yamlhelpers.dump_yaml(data)

    @classmethod
    def _load(cls, data: str, mode: yamlhelpers.LoaderStr = "unsafe") -> dict | list:
        return yamlhelpers.load_yaml(data, mode)


if __name__ == "__main__":
    info = YamlFile(".pre-commit-config.yaml")
    print(info)



from __future__ import annotations

import mknodes as mk


def test_empty():
    node = mk.MkText()
    assert not str(node)


def test_getitem_ending_with_eof():
    node = mk.MkText("## Test section\nTest")
    assert str(node["Test section"]) == "Test"


def test_getitem_ending_with_another_section():
    node = mk.MkText("## Test section\nTest\n## Another section")
    assert str(node["Test section"]) == "Test\n"


def test_fetch_from_url():
    url = "https://raw.githubusercontent.com/fire1ce/DDNS-Cloudflare-Bash/main/README.md"
    node = mk.MkText.from_url(url)
    assert str(node)



from __future__ import annotations

import markdown

from mknodes.utils import log


logger = log.get_logger(__name__)


class CustomFence:
    """Formatter wrapper."""

    def __init__(self, extensions: list):
        self.extensions = extensions

    def custom_formatter(
        self,
        source,
        language,
        css_class,
        options,
        md: markdown.Markdown,
        classes=None,
        id_value="",
        attrs=None,
        **kwargs,
    ):
        # import mknodes as mk

        # node_kls = getattr(mk, language)
        # print(f"{language=} {css_class=} {options=} {classes=} {id_value=} {attrs=}")
        try:
            return md.convert(source)
        except Exception:
            logger.exception("Error for custom fence %s", language)
            raise


def generate_fences():
    import mknodes as mk

    dcts = []
    for node_name in mk.__all__:
        dct = {"name": node_name, "class": node_name, "format": fence.custom_formatter}
        dcts.append(dct)
    return dcts


if __name__ == "__main__":
    fence = CustomFence(extensions=[])
    fence_dct = {"name": "ab", "class": "ab", "format": fence.custom_formatter}
    config = {"custom_fences": generate_fences()}
    md = markdown.Markdown(
        extensions=["pymdownx.superfences", "attr_list"],
        extension_configs={"pymdownx.superfences": config},
    )
    text = "```{.MkText shift_header_levels=1}\ntest\n```"
    result = md.convert(text)
    print(result)



from __future__ import annotations

from collections.abc import Callable
import pathlib
import types

from typing import Any

import fsspec
import fsspec.core
import jinja2

from mknodes.utils import pathhelpers, reprhelpers


class LoaderMixin:
    """Loader mixin which allows to OR loaders into a choice loader."""

    loader: jinja2.BaseLoader

    def __or__(self, other: jinja2.BaseLoader):
        own_loaders = self.loaders if isinstance(self, jinja2.ChoiceLoader) else [self]  # type: ignore[list-item]
        if isinstance(other, jinja2.ChoiceLoader):
            other_loaders = other.loaders
        else:
            other_loaders = [other]
        return ChoiceLoader([*own_loaders, *other_loaders])


class PackageLoader(LoaderMixin, jinja2.PackageLoader):
    """A loader for loading templates from a package."""

    def __init__(
        self,
        package: str | types.ModuleType,
        package_path: str | None = None,
        encoding: str = "utf-8",
    ) -> None:
        if isinstance(package, types.ModuleType):
            package = package.__name__
        parts = package.split(".")
        path = "/".join(parts[1:])
        if package_path:
            path = (pathlib.Path(path) / package_path).as_posix()
        super().__init__(parts[0], path, encoding)

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            package_name=self.package_name,
            package_path=self.package_path,
        )


class FileSystemLoader(LoaderMixin, jinja2.FileSystemLoader):
    """A loader to load templates from the file system."""

    def __repr__(self):
        return reprhelpers.get_repr(self, searchpath=self.searchpath)

    def __add__(self, other):
        if isinstance(other, jinja2.FileSystemLoader):
            paths = other.searchpath
        else:
            paths = [other]
        return FileSystemLoader([*self.searchpath, *paths])


class ChoiceLoader(LoaderMixin, jinja2.ChoiceLoader):
    """A loader which combines multiple other loaders."""

    def __repr__(self):
        return reprhelpers.get_repr(self, loaders=self.loaders, _shorten=False)


class DictLoader(LoaderMixin, jinja2.DictLoader):
    """A loader to load static content from a path->template-str mapping."""

    def __repr__(self):
        return reprhelpers.get_repr(self, mapping=self.mapping)

    def __add__(self, other):
        if isinstance(other, jinja2.DictLoader):
            mapping = self.mapping | other.mapping
        elif isinstance(other, dict):
            mapping = self.mapping | other
        return DictLoader(mapping)


class FsSpecProtocolPathLoader(LoaderMixin, jinja2.BaseLoader):
    """A jinja loader for fsspec filesystems.

    This loader allows to access templates from an fsspec protocol path,
    like "github://phil65:mknodes@main/README.md"

    Examples:
        ``` py
        loader = FsSpecProtocolPathLoader()
        env = Environment(loader=loader)
        env.get_template("github://phil65:mknodes@main/docs/icons.jinja").render()
        ```
    """

    def get_source(
        self,
        environment: jinja2.Environment,
        template: str,
    ) -> tuple[str, str, Callable[[], bool] | None]:
        src = pathhelpers.fsspec_get(template)
        path = pathlib.Path(template).as_posix()
        return src, path, lambda: True

    def list_templates(self) -> list[str]:
        return []

    def __repr__(self):
        return reprhelpers.get_repr(self)


class FsSpecFileSystemLoader(LoaderMixin, jinja2.BaseLoader):
    """A jinja loader for fsspec filesystems.

    This loader allows to access templates from an fsspec filesystem.

    Template paths must be relative to the filesystem root.
    In order to access templates via protocol path, see `FsSpecProtocolPathLoader`.

    Examples:
        ``` py
        # protocol path
        loader = FsSpecFileSystemLoader("dir::github://phil65:mknodes@main/docs")
        env = Environment(loader=loader)
        env.get_template("icons.jinja").render()

        # protocol and storage options
        loader = FsSpecFileSystemLoader("github", org="phil65", repo="mknodes")
        env = Environment(loader=loader)
        env.get_template("docs/icons.jinja").render()

        # fsspec filesystem
        fs = fsspec.filesystem("github", org="phil65", repo="mknodes")
        loader = FsSpecFileSystemLoader(fs)
        env = Environment(loader=loader)
        env.get_template("docs/icons.jinja").render()
        ```

    """

    def __init__(self, fs: fsspec.AbstractFileSystem | str, **kwargs: Any):
        """Constructor.

        Arguments:
            fs: Either a protocol path string or an fsspec filesystem instance.
                Also supports "::dir" prefix to set the root path.
            kwargs: Optional storage options for the filesystem.
        """
        super().__init__()
        if isinstance(fs, str):
            if "://" in fs:
                self.fs, self.path = fsspec.core.url_to_fs(fs)
            else:
                self.fs = fsspec.filesystem(fs, **kwargs)
        else:
            self.fs = fsspec.filesystem(fs, **kwargs) if isinstance(fs, str) else fs
            self.path = ""

    def __repr__(self):
        return reprhelpers.get_repr(self, fs=self.fs.protocol)

    def list_templates(self) -> list[str]:
        return self.fs.ls("")

    def get_source(
        self,
        environment: jinja2.Environment,
        template: str,
    ) -> tuple[str, str, Callable[[], bool] | None]:
        with self.fs.open(template) as file:
            src = file.read().decode()

        path = pathlib.Path(template).as_posix()
        return src, path, lambda: True


def get_loader(
    module_paths: list[str] | None = None,
    file_paths: list[str] | None = None,
    static: dict[str, str] | None = None,
    fsspec_paths: bool = True,
) -> ChoiceLoader:
    loaders: list[jinja2.BaseLoader] = [PackageLoader(p) for p in module_paths or []]
    if file_paths:
        loaders.append(FileSystemLoader(file_paths))
    if static:
        loaders.append(DictLoader(static))
    if fsspec_paths:
        loaders.append(FsSpecProtocolPathLoader())
    return ChoiceLoader(loaders)


resources_loader = PackageLoader("mknodes", "resources")
docs_loader = FileSystemLoader(searchpath="docs/")
fsspec_protocol_loader = FsSpecProtocolPathLoader()
resource_loader = ChoiceLoader([resources_loader, docs_loader, fsspec_protocol_loader])


LOADERS = dict(
    fsspec=FsSpecFileSystemLoader,
    filesystem=FileSystemLoader,
    package=PackageLoader,
    dictionary=DictLoader,
)


if __name__ == "__main__":
    from mknodes.jinja import environment

    loader = FsSpecFileSystemLoader("dir::github://phil65:mknodes@main/docs")
    env = environment.Environment()
    env.loader = loader
    template = env.get_template("icons.jinja")
    print(template.render())
    loader = FsSpecProtocolPathLoader()
    result = loader.get_source(env, "github://phil65:mknodes@main/README.md")
    print(repr(loader))



from __future__ import annotations

from mknodes.utils import log
import re
import textwrap

from mknodes.pages.metadata import Metadata


HEADER_REGEX = re.compile(r"^(#{1,6}) (.*)", flags=re.MULTILINE)

logger = log.get_logger(__name__)


class TextProcessor:
    ID: str

    def run(self, text: str) -> str:
        raise NotImplementedError


class AnnotationProcessor(TextProcessor):
    ID = "annotations"

    def __init__(self, item):
        self.item = item

    def run(self, text: str) -> str:
        return self.item.attach_annotations(text) if self.item.annotations else text


class FootNotesProcessor(TextProcessor):
    ID = "footnotes"

    def __init__(self, item):
        self.item = item

    def run(self, text: str) -> str:
        return f"{text}\n\n{self.item.footnotes}" if self.item.footnotes else text


class IndentationProcessor(TextProcessor):
    ID = "indentation"

    def __init__(self, indent: int | str = 4):
        self.indent = " " * indent if isinstance(indent, int) else indent

    def run(self, text: str) -> str:
        return textwrap.indent(text, self.indent) if self.indent else text


class ShiftHeaderLevelProcessor(TextProcessor):
    ID = "shift_header_levels"

    def __init__(self, level_shift):
        self.level_shift = level_shift

    def run(self, text: str) -> str:
        if not self.level_shift:
            return text

        def mod_header(match: re.Match, levels: int) -> str:
            header_str = match[1]
            if levels > 0:
                header_str += levels * "#"
            else:
                header_str = header_str[:levels]
            return f"{header_str} {match[2]}"

        return re.sub(HEADER_REGEX, lambda x: mod_header(x, self.level_shift), text)


class RenderJinjaProcessor(TextProcessor):
    ID = "render_jinja_templates"

    def __init__(self, env, variables=None):
        self.env = env
        self.variables = variables or {}

    def run(self, text: str) -> str:
        return self.env.render(text, self.variables)


class AppendCssClassesProcessor(TextProcessor):
    ID = "append_css_classes"

    def __init__(self, item):
        self.item = item

    def run(self, text: str) -> str:
        return self.item.attach_css_classes(text) if self.item.mods.css_classes else text


class PrependMetadataProcessor(TextProcessor):
    ID = "prepend_metadata"

    def __init__(self, metadata: Metadata):
        self.metadata = metadata

    def run(self, text: str) -> str:
        header = self.metadata.as_page_header()
        return f"{header}\n{text}" if header else text


class PrependHeaderProcessor(TextProcessor):
    ID = "prepend_header"

    def __init__(self, header: str | None):
        self.header = header

    def run(self, text: str) -> str:
        if not self.header:
            return text
        header = self.header if self.header.startswith("#") else f"## {self.header}"
        return f"{header}\n\n{text}"


if __name__ == "__main__":
    text = "## test"
    processor = ShiftHeaderLevelProcessor(3)
    text = processor.run(text)
    print(text)



from __future__ import annotations

import abc

from collections.abc import Collection, Mapping
import io
import itertools
import os
import pathlib
import posixpath
import re
import types
from typing import BinaryIO, Self
import zlib

from mknodes.utils import downloadhelpers, helpers, log


logger = log.get_logger(__name__)


INV_HEADER = """\
# Sphinx inventory version 2
# Project: {project}
# Version: {version}
# The remainder of this file is compressed using zlib.
"""


class InventoryItem:
    """Inventory item."""

    def __init__(
        self,
        name: str,
        domain: str,
        role: str,
        uri: str,
        priority: int = 1,
        dispname: str | None = None,
    ):
        """Initialize the object.

        Arguments:
            name: The item name.
            domain: The item domain, like 'python'
            role: The item role, like 'class' or 'method'.
            uri: The item URI.
            priority: The item priority.
            dispname: The item display name.
        """
        self.name: str = name
        self.domain: str = domain
        self.role: str = role
        self.uri: str = uri
        self.priority: int = priority
        self.dispname: str = dispname or name

    def format_sphinx(self) -> str:
        """Format this item as a Sphinx inventory line and return it."""
        dispname = self.dispname
        if dispname == self.name:
            dispname = "-"
        uri = self.uri
        if uri.endswith(self.name):
            uri = uri[: -len(self.name)] + "$"
        return f"{self.name} {self.domain}:{self.role} {self.priority} {uri} {dispname}"

    sphinx_item_regex = re.compile(r"^(.+?)\s+(\S+):(\S+)\s+(-?\d+)\s+(\S+)\s*(.*)$")

    @classmethod
    def parse_sphinx(cls, line: str) -> InventoryItem:
        """Parse a line from a Sphinx v2 inventory file and return an `InventoryItem`.

        Arguments:
            line: The line to parse
        """
        match = cls.sphinx_item_regex.search(line)
        if not match:
            raise ValueError(line)
        name, domain, role, priority, uri, dispname = match.groups()
        if uri.endswith("$"):
            uri = uri[:-1] + name
        if dispname == "-":
            dispname = name
        return cls(name, domain, role, uri, int(priority), dispname)


class BaseInventory(dict):
    """Inventory of collected and rendered objects."""

    def __init__(
        self,
        items: list[InventoryItem] | None = None,
        project: str = "project",
        version: str = "0.0.0",
    ):
        """Initialize the object.

        Arguments:
            items: A list of items.
            project: The project name.
            version: The project version.
        """
        super().__init__()
        items = items or []
        for item in items:
            self[item.name] = item
        self.project = project
        self.version = version

    def register(
        self,
        name: str,
        domain: str,
        role: str,
        uri: str,
        priority: int = 1,
        dispname: str | None = None,
    ) -> None:
        """Create and register an item.

        Arguments:
            name: The item name.
            domain: The item domain, like 'python'
            role: The item role, like 'class' or 'method'.
            uri: The item URI.
            priority: The item priority.
            dispname: The item display name.
        """
        self[name] = InventoryItem(
            name=name,
            domain=domain,
            role=role,
            uri=uri,
            priority=priority,
            dispname=dispname,
        )

    def format_sphinx(self) -> bytes:
        """Format this inventory as a Sphinx `objects.inv` file and return it."""
        header = INV_HEADER.format(project=self.project, version=self.version).encode()
        lines = [
            item.format_sphinx().encode()
            for item in sorted(self.values(), key=lambda item: (item.domain, item.name))
        ]
        return header + zlib.compress(b"\n".join(lines) + b"\n", 9)

    @classmethod
    def parse_sphinx(
        cls,
        in_file: BinaryIO,
        *,
        domain_filter: Collection[str] = (),
    ) -> Self:
        """Parse a Sphinx v2 inventory file and return an `Inventory` from it.

        Arguments:
            in_file: The binary file-like object to read from.
            domain_filter: A collection of domain values to allow.
        """
        for _ in range(4):
            in_file.readline()
        lines = zlib.decompress(in_file.read()).splitlines()
        items = [InventoryItem.parse_sphinx(line.decode()) for line in lines]
        if domain_filter:
            items = [item for item in items if item.domain in domain_filter]
        return cls(items)


class Inventory(BaseInventory):
    def __init__(self, base_url: str):
        self.base_url = base_url

    @classmethod
    def from_file(
        cls,
        path: str | os.PathLike | io.BytesIO,
        base_url: str,
        *,
        domains: list[str] | None = None,
    ):  # sourcery skip: assign-if-exp
        inv = cls(base_url)
        domains = domains or ["py"]
        if isinstance(path, io.BytesIO):  # noqa: SIM108
            file = path
        else:
            file = pathlib.Path(path).open("rb")  # noqa: SIM115
        with file:
            try:
                inv_dict = BaseInventory.parse_sphinx(file, domain_filter=domains)
            except zlib.error as e:
                logger.warning("Error when parsing Inventory file: %s", e)
                return inv
        inv.update(inv_dict)
        return inv

    @classmethod
    def from_url(
        cls,
        url: str,
        *,
        base_url: str | None = None,
        domains: list[str] | None = None,
    ):
        """Return an Inventory based on an inventory file located at given url.

        Arguments:
            url: Inventory file url
            base_url: The base url for the inventory, if different from download url
            domains: The domains to include
        """
        data = downloadhelpers.download(url)
        buffer = io.BytesIO(data)
        if base_url is None:
            base_url = os.path.dirname(url)  # noqa: PTH120
        return cls.from_file(buffer, base_url or "", domains=domains)

    def __getitem__(self, value):
        val = super().__getitem__(value)
        return posixpath.join(self.base_url, val.uri)


class InventoryManager(Mapping, metaclass=abc.ABCMeta):
    # TODO: might be worth using collections.ChainMap, or just merging all inv files.

    def __init__(self):
        self.inv_files: list[BaseInventory] = []

    def add_inv_file(
        self,
        path: str | os.PathLike,
        base_url: str | None = None,
        domains: list[str] | None = None,
    ):
        import urllib.error

        path = str(path)
        if helpers.is_url(path):
            logger.debug("Downloading %r...", path)
            try:
                inv = Inventory.from_url(path, base_url=base_url, domains=domains)
                self.inv_files.append(inv)
            except urllib.error.HTTPError:
                logger.debug("No file for %r...", path)
                return
        elif base_url:
            inv = Inventory.from_file(path, domains=domains, base_url=base_url)
            self.inv_files.append(inv)
        else:
            msg = "Base URL needed for loading from file."
            raise ValueError(msg)

    def __getitem__(self, name: str | type | types.FunctionType | types.MethodType):
        match name:
            case type():
                path = f"{name.__module__}.{name.__qualname__}"
            case str():
                path = name
            case _:
                raise TypeError(name)
        for inv_file in self.inv_files:
            if path in inv_file:
                return inv_file[path]
        raise KeyError(name)

    def __iter__(self):
        return itertools.chain(*[inv_file.keys() for inv_file in self.inv_files])

    def __len__(self):
        return sum(len(i) for i in self.inv_files)


if __name__ == "__main__":
    inv_manager = InventoryManager()
    inv_manager.add_inv_file("tests/data/objects.inv", base_url="http://test.de")
    file = inv_manager.inv_files[0]
    item = inv_manager["prettyqt.widgets.widget.WidgetMixin.set_style"]
    # print(item)
    inv = Inventory.from_url("https://docs.python.org/3/objects.inv")
    print(inv)
    # print(inv_manager["prettyqt.widgets.widget.WidgetMixin.set_style"])
    # print(file)



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mknode
from mknodes.utils import log, reprhelpers


if TYPE_CHECKING:
    import mknodes as mk


logger = log.get_logger(__name__)


class MkCallable(mknode.MkNode):
    """Node carrying a callable. Can be either page, nav or node.

    The callable is called each time the virtual files are requested or when
    a Markdown conversion is requested.

    Experimental node. I do not really know yet what this should be good for. :)
    """

    ICON = "material/function"
    STATUS = "new"

    def __init__(
        self,
        fn: Callable[..., mk.MkNode],
        *,
        args: list | tuple | None = None,
        kw_args: dict | None = None,
        header: str = "",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            fn: Callable (must return a MkNode).
            args: Arguments to use for callable
            kw_args: Keyword arguments to use for callable
            header: Section header
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(header=header, **kwargs)
        self.fn = fn
        self.args = args or []
        self.kw_args = kw_args or {}

    def __call__(self) -> mk.MkNode:
        node = self.fn(*self.args, **self.kw_args)
        node.parent = self.parent
        return node

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            fn=self.fn,
            args=self.args,
            kw_args=self.kw_args,
            _filter_empty=True,
        )

    @property
    def files(self):
        node = self.__call__()
        return node.files

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkCallable(lambda: mk.MkAdmonition("Nested!"))
        page += mk.MkReprRawRendered(node)

    def to_markdown(self) -> str:
        node = self.__call__()
        return node.to_markdown()


if __name__ == "__main__":
    import mknodes as mk

    def make_page():
        page = mk.MkPage("test", hide="toc", inclusion_level=False)
        page += "Some content"
        return page

    text = MkCallable(make_page, header="test")
    print(text.to_markdown())



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from mknodes.basenodes import mkcontainer, mknode, mktabs, mktext
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class MkTabContainer(mkcontainer.MkContainer):
    """Base class for nodes containing tabs."""

    items: list[mktabs.MkTab | mktabs.MkTabBlock]
    Tab: type[mktabs.MkTab | mktabs.MkTabBlock]
    ICON = "material/tab"

    def __init__(
        self,
        tabs: Mapping[str, str | mknode.MkNode | list] | list[mktabs.MkTab],
        *,
        header: str = "",
        select_tab: int | str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            tabs: Tab data
            header: Section header
            select_tab: Tab which should be selected initially
            kwargs: Keyword arguments passed to parent
        """
        if isinstance(tabs, list):
            items = tabs
        else:
            items = [
                self.Tab(
                    title=k,
                    content=mktext.MkText(v) if isinstance(v, str) else v,
                )
                for k, v in tabs.items()
            ]
        self.select_tab = select_tab
        super().__init__(content=items, header=header, **kwargs)

    def __getitem__(self, item: int | str) -> mktabs.MkTab | mktabs.MkTabBlock:
        match item:
            case int():
                return self.items[item]
            case str():
                for tab in self.items:
                    if tab.title == item:
                        return tab
                raise IndexError(item)
            case _:
                raise TypeError(item)

    def __contains__(self, tab: str | mktabs.MkTab | mktabs.MkTabBlock) -> bool:
        match tab:
            case mktabs.MkTab() | mktabs.MkTabBlock():
                return tab in self.items
            case str():
                return any(i.title == tab for i in self.items)
            case _:
                raise TypeError(tab)

    def add_tab(
        self,
        title: str,
        content: str | mknode.MkNode | list[str] | list[mknode.MkNode],
        select: bool = False,
    ) -> mktabs.MkTab | mktabs.MkTabBlock:
        """Append a tab to existing ones.

        Arguments:
            title: Title of the new tab
            content: content of the new tab
            select: Whether new tab should get selected initially
        """
        tab = self.Tab(title, content, select=select)
        self.append(tab)
        return tab

    def _get_tab_pos(self, tab_title: str) -> int:
        item = next(i for i in self.items if i.title == tab_title)
        return self.items.index(item)

    def set_selected(self, index: int | str):
        """Set tab with given index as selected.

        Arguments:
            index: Index or title of the tab which should be selected
        """
        self.select_tab = self._get_tab_pos(index) if isinstance(index, str) else index
        for i, item in enumerate(self.items):
            item.select = i == self.select_tab

    def __setitem__(self, index: str, value: mktabs.MkTab | mktabs.MkTabBlock | str):
        match value:
            case str():
                item = mktext.MkText(value)
                tab = self.Tab(index, content=item)
            case mktabs.MkTab() | mktabs.MkTabBlock():
                tab = value
            case mknode.MkNode():
                tab = self.Tab(index, content=value)
        if index in self:
            pos = self._get_tab_pos(index)
            self.items[pos] = tab
        else:
            self.items.append(tab)

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            tabs=self.to_dict(),
            select_tab=self.select_tab,
            _filter_empty=True,
        )

    def to_dict(self) -> dict[str, str]:
        return {tab.title: "\n\n".join(str(i) for i in tab.items) for tab in self.items}

    def _to_markdown(self) -> str:
        if not self.items:
            return ""
        if self.select_tab is not None:
            self.set_selected(self.select_tab)
        self.items[0].new = True
        return "\n".join(str(i) for i in self.items)


if __name__ == "__main__":
    tabs = dict(Tab1="Some text", Tab2="Another text")
    tabblock = MkTabContainer(tabs)
    print(tabblock)



from __future__ import annotations as _annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mkcontainer, mktext
from mknodes.utils import log, reprhelpers, resources


if TYPE_CHECKING:
    import mknodes as mk


logger = log.get_logger(__name__)


class MkAnnotation(mkcontainer.MkContainer):
    """Represents a single annotation. It gets managed by an MkAnnotations node."""

    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.superfences")]

    def __init__(
        self,
        num: int,
        content: str | mk.MkNode,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            num: Annotation index number
            content: Annotation content
            kwargs: Keyword arguments passed to parent
        """
        self.num = num
        super().__init__(content=content, **kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self, num=self.num, content=self.items)

    def _to_markdown(self) -> str:
        item_str = "\n\n".join(i.to_markdown() for i in self.items)
        lines = item_str.split("\n")
        space = (3 - len(str(self.num))) * " "
        result = [f"{self.num}.{space}{lines[0]}"]
        result.extend(f"    {i}" for i in lines[1:])
        return "\n".join(result) + "\n"


class MkAnnotations(mkcontainer.MkContainer):
    """Node containing a list of MkAnnotations."""

    items: list[MkAnnotation]
    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.superfences")]
    ICON = "material/alert-box"

    def __init__(
        self,
        annotations: (
            Mapping[int, str | mk.MkNode] | list[MkAnnotation] | list[str] | None
        ) = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            annotations: Annotations data (Can be given in different shapes)
            kwargs: Keyword arguments passed to parent
        """
        match annotations:
            case None:
                items = []
            case list():
                items = [
                    (
                        ann
                        if isinstance(ann, MkAnnotation)
                        else MkAnnotation(i, ann)  # type: ignore
                    )
                    for i, ann in enumerate(annotations, start=1)
                ]
            case Mapping():
                items = [
                    MkAnnotation(
                        k,
                        content=mktext.MkText(v) if isinstance(v, str) else v,
                    )
                    for k, v in annotations.items()
                ]
            case _:
                raise TypeError(annotations)
        super().__init__(content=items, **kwargs)

    def __len__(self):
        return len(self.items)

    def __getitem__(self, item: int):
        for node in self.items:
            if node.num == item:
                return node
        raise IndexError(item)

    def __contains__(self, item: int | MkAnnotation) -> bool:
        match item:
            case MkAnnotation():
                return item in self.items
            case int():
                return any(i.num == item for i in self.items)
            case _:
                raise TypeError(item)

    def __repr__(self):
        notes = []
        for item in self.items:
            if len(item.items) == 1:
                item = reprhelpers.to_str_if_textnode(item.items[0])
            notes.append(item)
        return reprhelpers.get_repr(self, annotations=notes)

    def _get_item_pos(self, num: int) -> int:
        item = next(i for i in self.items if i.num == num)
        return self.items.index(item)

    def __setitem__(self, index: int, value: mk.MkNode | str):
        import mknodes as mk

        match value:
            case str():
                item = mk.MkText(value)
                node = MkAnnotation(index, content=item, parent=self)
            case MkAnnotation():
                node = value
            case mk.MkNode():
                node = MkAnnotation(index, content=value, parent=self)
        if index in self:
            pos = self._get_item_pos(index)
            self.items[pos] = node
        else:
            self.items.append(node)

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        page += mk.MkCode.for_object(
            MkAnnotations.create_example_page,
            extract_body=True,
        )
        node = MkAnnotations()
        page += node
        text = mk.MkText("The MkAnnotations node aggregates annotations(1).")
        page += text
        info = r"Annotations are numbered and can be set via \__setitem__."
        node[1] = info  # (1)
        admonition = mk.MkAdmonition("They can also contain other Markdown.")
        node[2] = admonition  # (2)
        text.annotations[1] = "Every MkNode can annotate via the 'annotations' attribute"
        page += mk.MkCode(str(node), language="markdown", header="Markdown")

    def _to_markdown(self) -> str:
        if not self.items:
            return ""
        items = sorted(self.items, key=lambda x: x.num)
        return "".join(i.to_markdown() for i in items)

    def annotate_text(self, markdown: str) -> str:
        if not self.items:
            return markdown
        return f'<div class="annotate" markdown>\n{markdown}\n</div>\n\n{self}'


if __name__ == "__main__":
    import mknodes as mk

    page = mk.MkPage()
    MkAnnotations.create_example_page(page)
    print(page)



from __future__ import annotations

from mknodes.navs import mknav
from mknodes.utils import log


logger = log.get_logger(__name__)


class MkEmptyWebsite(mknav.MkNav):
    """Non-populated MkNav which parses given static pages dict."""

    def __init__(self, static_pages: dict | None = None, **kwargs):
        super().__init__(**kwargs)
        import mknodes as mk

        page = self.add_page(is_index=True, title="Overview", hide="toc")
        page += mk.MkText(r"metadata.description", is_jinja_expression=True)
        static_pages = static_pages or {}
        self.parse.json(static_pages)

    @classmethod
    def for_project(cls, project, **kwargs):
        root = cls(context=project.context, **kwargs)
        project.set_root(root)
        return root


if __name__ == "__main__":
    import mknodes as mk

    theme = mk.MaterialTheme()
    doc = MkEmptyWebsite.for_project(
        mk.Project(theme=theme),
        static_pages={
            "Usage": "https://raw.githubusercontent.com/mkdocs/mkdocs/master/docs/getting-started.md",
        },
    )
    print(doc)



from __future__ import annotations

import dataclasses
import pathlib
import types

from typing import Any

from griffe.dataclasses import Alias, Module

import mknodes as mk

from mknodes.data import buildsystems, commitconventions, installmethods, tools
from mknodes.info import (
    folderinfo,
    linkprovider,
    mkdocsconfigfile,
    packageregistry,
    pyproject,
)
from mknodes.utils import clihelpers, log, superdict


logger = log.get_logger(__name__)


@dataclasses.dataclass
class Context:
    """Base class for contexts."""

    def as_dict(self):
        return {
            field.name: getattr(self, field.name) for field in dataclasses.fields(self)
        }

    @property
    def fields(self) -> list[str]:
        return [i.name for i in dataclasses.fields(self)]


@dataclasses.dataclass
class GitContext(Context):
    """Local repository information."""

    main_branch: str = ""
    """Name of the main branch of the repo (`master` / `main`)."""
    repo_name: str = ""
    """Name of the git folder."""
    last_commits: list = dataclasses.field(default_factory=list)
    """List of last commits (Commit objects from `GitPython`)."""
    repo_hoster: str = ""
    """Name of the code hoster (for example `GitHub`)"""
    edit_uri: str | None = None
    """Edit uri (depends on code hoster)."""


@dataclasses.dataclass
class ThemeContext(Context):
    """Information about the theme."""

    name: str = ""
    """Name of the theme."""
    primary_color: str = ""
    """Primary color."""
    text_color: str = ""
    """Primary text color."""
    data: dict[str, str] = dataclasses.field(default_factory=dict)
    """Additional data of the theme."""
    admonitions: list = dataclasses.field(default_factory=list)
    css_primary_fg: str = ""
    css_primary_bg: str = ""
    css_primary_bg_light: str = ""
    css_accent_fg: str = ""
    css_accent_fg_transparent: str = ""
    css_accent_bg: str = ""
    css_default_fg: str = ""
    css_default_bg: str = ""


# @dataclasses.dataclass
# class PyProjectContext(Context):
#     # pyproject
#     build_system: buildsystems.BuildSystem = dataclasses.field(
#         default_factory=lambda: buildsystems.hatch,
#     )
#     """The build system set as build backend *[pyproject]*"""
#     configured_build_systems: list = dataclasses.field(default_factory=list)
#     """A list of build systems which are configured in pyproject *[pyproject]*"""
#     tool_section: dict[str, Any] = dataclasses.field(default_factory=dict)
#     """The tool section of the pyproject file (as a dict) *[pyproject]*"""
#     line_length: int | None = None
#     """The line length, if defined by any popular tool *[pyproject]*"""
#     commit_types: list[commitconventions.CommitTypeStr] = dataclasses.field(
#         default_factory=list,
#     )
#     """Commit types defined in pyproject mknodes section *[pyproject]*"""
#     extras_descriptions: dict[str, str] = dataclasses.field(default_factory=dict)
#     """Descriptions for extras, defined in pyproject mknodes section *[pyproject]*"""
#     package_repos: list[installmethods.InstallMethodStr] = dataclasses.field(
#         default_factory=list,
#     )
#     """Package repositories the distribution is distributed on."""
#     docstring_style: str | None = None
#     """Defined in pyproject mknodes section *[pyproject]*"""


@dataclasses.dataclass
class PackageContext(Context):
    """Information about a package."""

    # PackageInfo
    module: types.ModuleType | None = None
    """The module object itself."""
    griffe_module: Module | Alias | None = None
    """The module object itself."""
    docstring_style: str | None = None
    """The style used for DocStrings."""
    distribution_name: str = ""
    """The name of the distribution *[Metadata]*"""
    version: str = ""
    """The version of the distribution *[Metadata]*"""
    summary: str = ""
    """A summary for the distribution *[Metadata]*"""
    description: str = ""
    """A description for the distribution *[Metadata]*"""
    author_name: str = ""
    """The author name of the distribution *[Metadata]*"""
    author_email: str = ""
    """A description for the distribution *[Metadata]*"""
    authors: dict[str, str] = dataclasses.field(default_factory=dict)
    """All authors of the distribution *[Metadata]*"""
    classifiers: list = dataclasses.field(default_factory=list)
    """Distribution classifiers *[Metadata]*"""
    classifier_map: dict[str, list[str]] = dataclasses.field(default_factory=dict)
    """Distribution classifiers, sorted by category *[Metadata]*"""
    keywords: list[str] = dataclasses.field(default_factory=list)
    """Distribution keywords *[Metadata]*"""
    required_python_version: str | None = ""
    """The required python version for the distribution *[Metadata]*"""
    required_package_names: list[str] = dataclasses.field(default_factory=list)
    """The names of the dependencies."""
    required_packages: dict = dataclasses.field(default_factory=dict)
    """PackageInfos for the dependencies."""
    extras: dict[str, folderinfo.PackageExtra] = dataclasses.field(default_factory=dict)
    """The extras of the distribution."""
    urls: dict[str, str] = dataclasses.field(default_factory=dict)
    """A set of URLs related to the distribution."""
    homepage: str = ""
    """The main website of the distribution."""
    license_name: str | None = ""
    """Name of the license"""
    license_text: str | None = ""
    """The complete license text"""
    pretty_name: str = ""
    """A pretty name for the distribution (like **MkNodes**) *[MkDocs site name]*"""
    repository_url: str = ""
    """The URL to the remote code repository."""
    repository_username: str = ""
    """The username for the remote code repository."""
    repository_name: str = ""
    """The repository name of the distribution."""
    repository_path: pathlib.Path = dataclasses.field(default_factory=pathlib.Path)
    """The path to the local git repository."""
    mkdocs_config: mkdocsconfigfile.MkDocsConfigFile | None = None
    """A dict-like File containing the MkDocs config."""
    pyproject_file: pyproject.PyProject = dataclasses.field(
        default_factory=pyproject.PyProject,
    )
    """A dict-like File containing the PyProject data."""
    tools: list[tools.Tool] = dataclasses.field(default_factory=list)
    """A list of tools found for the distribution."""
    task_runners: list = dataclasses.field(default_factory=list)
    """Task runners used by the distribution."""
    social_info: list[dict[str, str]] = dataclasses.field(default_factory=list)
    """A icon-name -> URL dictionary containing ."""
    inventory_url: str | None = ""
    """A best guess for an inventory URL for the package."""
    entry_points: dict = dataclasses.field(default_factory=dict)
    """A dictionary containing the entry points of the distribution."""
    cli: str | None = None
    """The cli package name used by the distribution."""
    cli_info: clihelpers.CommandInfo | None = None
    """An object containing information about all cli commands."""

    # required_packages: dict[PackageInfo, packagehelpers.Dependency] =

    # pyproject
    build_system: buildsystems.BuildSystem = dataclasses.field(
        default_factory=lambda: buildsystems.hatch,
    )
    """The build system set as build backend *[pyproject]*"""
    configured_build_systems: list = dataclasses.field(default_factory=list)
    """A list of build systems which are configured in pyproject *[pyproject]*"""
    tool_section: superdict.SuperDict[Any] = dataclasses.field(
        default_factory=superdict.SuperDict,
    )
    """The tool section of the pyproject file.

    Comes as a SuperDict. (A Mutable mapping with extended capabilities) *[pyproject]*"""
    line_length: int | None = None
    """The line length, if defined by any popular tool *[pyproject]*"""
    commit_types: list[commitconventions.CommitTypeStr] = dataclasses.field(
        default_factory=list,
    )
    """Commit types defined in pyproject mknodes section *[pyproject]*"""
    package_repos: list[installmethods.InstallMethodStr] = dataclasses.field(
        default_factory=list,
    )
    """Package repositories the distribution is distributed on.
    Defined in pyproject mknodes section *[pyproject]*"""


@dataclasses.dataclass
class GitHubContext(Context):
    """Information about the GitHub repo / user."""

    default_branch: str = ""
    """The default branch of the repository."""
    repo_name: str = ""
    """The repository name."""
    workflows: list[dict] = dataclasses.field(default_factory=list)
    """A dictionary (workflow-name -> workflow-yaml) containing workflows."""
    avatar_url: str | None = None
    """The url of the GitHub avatar."""
    bio: str | None = None
    """The user biography."""
    blog: str | None = None
    """The user blog URL."""
    company: str | None = None
    """The company associated with the user."""
    email: str | None = None
    """The user email address."""
    followers: int = 0
    """The follower count of the user."""
    gravatar_id: str | None = None
    """The gravatar id associated with the user."""
    hireable: bool = False
    """Whether the user is hireable."""
    location: str | None = None
    """The user location, as set on GitHub."""
    name: str | None = None
    """The GitHub username."""
    twitter_username: str | None = None
    """The twitter username."""


@dataclasses.dataclass
class ProjectContext(Context):
    """All information about a project."""

    metadata: PackageContext = dataclasses.field(default_factory=PackageContext)
    """Information about a package."""
    git: GitContext = dataclasses.field(default_factory=GitContext)
    """Local repository information."""
    github: GitHubContext = dataclasses.field(default_factory=GitHubContext)
    """Information about the GitHub repo / user."""
    theme: ThemeContext = dataclasses.field(default_factory=ThemeContext)
    """Information about the theme."""
    links: linkprovider.LinkProvider = dataclasses.field(
        default_factory=linkprovider.LinkProvider,
    )
    """Link source."""
    env: mk.Environment = dataclasses.field(
        default_factory=lambda: mk.Environment(load_templates=True),
    )
    # resources: Resources = dataclasses.field(default_factory=Resources)
    # pyproject: pyproject.PyProject = dataclasses.field(
    #     default_factory=pyproject.PyProject,
    # )

    def __post_init__(self):
        self.env.filters["get_link"] = self.links.get_link
        self.env.filters["get_url"] = self.links.get_url
        self.env.globals |= self.as_dict()

    def populate_linkprovider(self):
        if self.metadata.mkdocs_config is None:
            return
        invs = self.metadata.mkdocs_config.get_inventory_infos()
        mk_urls = {i["url"]: i.get("base_url") for i in invs if "url" in i}
        for url, base_url in mk_urls.items():
            self.links.add_inv_file(url, base_url=base_url)
        for url in packageregistry.registry.inventory_urls:
            if url not in mk_urls:
                self.links.add_inv_file(url)

    def as_dict(self):
        return dict(
            metadata=self.metadata,
            git=self.git,
            github=self.github,
            # resources=dict(self.resources),
            theme=self.theme,
            links=self.links,
        )


default_package_context = PackageContext(
    pretty_name="MkNodes",
    version="1.0.0",
    distribution_name="mknodes",
    summary="Don't write docs. Code them.",
    description="Long text with description.",
    author_name="Philipp Temminghoff",
    author_email="philipptemminghoff@gmail.com",
    license_name="MIT",
    required_python_version=">= 3.11",
    homepage="",
    cli="typer",
    module=mk,
    repository_url="https://github.com/phil65/mknodes/",
    repository_username="phil65",
    repository_name="mknodes",
    inventory_url="https://github.io/phil65/mknodes/objects.inv",
)


default_git_context = GitContext(
    main_branch="main",
    repo_name="mknodes",
    repo_hoster="GitHub",
    edit_uri="edit/main/docs/",
)


default_theme_context = ThemeContext(
    name="material",
    primary_color="AAAAAA",
    text_color="#000000",
)


default_github_context = GitHubContext(
    default_branch="main",
    repo_name="mknodes",
)


def get_default_project_context():
    return ProjectContext(
        metadata=default_package_context,
        git=default_git_context,
        github=default_github_context,
        theme=default_theme_context,
        links=linkprovider.LinkProvider(),
        env=mk.Environment(load_templates=True),
        # resources=Resources(),
    )


if __name__ == "__main__":
    info = PackageContext()
    print(info.fields)



from __future__ import annotations

import json

from mknodes.info import configfile


class JsonFile(configfile.ConfigFile):
    @classmethod
    def _dump(cls, data: dict) -> str:
        return json.dumps(data)

    @classmethod
    def _load(cls, data: str) -> dict | list:
        return json.loads(data)


if __name__ == "__main__":
    info = JsonFile("pyproject.toml")
    text = info.get_section_text("tool", "hatch", keep_path=True)
    print(text)



from __future__ import annotations

from collections.abc import Callable, Iterator, Mapping, Sequence
from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mkcontainer
from mknodes.utils import log, reprhelpers, resources


if TYPE_CHECKING:
    import mknodes as mk

logger = log.get_logger(__name__)


class MkBaseTable(mkcontainer.MkContainer):
    """Base Class for MkTables. Only deals with managing the data.

    Subclasses can use other mechanisms for the rendering, like external libraries.
    """

    REQUIRED_EXTENSIONS = [resources.Extension("tables")]
    ICON = "octicons/table-24"

    def __init__(
        self,
        data: Sequence[Sequence[str]] | Sequence[dict] | Mapping[str, list] | None = None,
        columns: Sequence[str] | None = None,
        *,
        header: str = "",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            data: Data show for the table
            columns: Column headers if not provided by data.
            header: Section header
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(header=header, **kwargs)
        match data:
            case () | None:
                self._data: dict[str, list[mk.MkNode]] = {c: [] for c in columns or []}
            case Mapping():
                self._data = {
                    str(k): [self.to_child_node(i) for i in v] for k, v in data.items()
                }
            case ((str(), *_), *_):
                h = columns or [str(i) for i in range(len(data))]
                self._data = {}
                for i, col in enumerate(data):
                    self._data[h[i]] = [self.to_child_node(j) for j in col]
            case (dict(), *_):
                self._data = {
                    k: [self.to_child_node(dic[k]) for dic in data]  # type: ignore[index]
                    for k in data[0]
                }
            case _:
                raise TypeError(data)

    def __repr__(self):
        kwarg_data = {
            k: [reprhelpers.to_str_if_textnode(i) for i in v]
            for k, v in self.data.items()
        }
        return reprhelpers.get_repr(self, data=kwarg_data)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def columns(self):
        return list(self.data.keys())

    @property
    def items(self):
        data = self.data  # property
        return [i for k in data for i in data[k]]

    @items.setter
    def items(self, data):
        match data:
            case Mapping():
                self._data = {
                    str(k): [self.to_child_node(i) for i in v] for k, v in data.items()
                }
            case (str(), *_):
                self._data = {"": [self.to_child_node(i) for i in data]}
            case (dict(), *_):
                self._data = {
                    k: [self.to_child_node(dic[k]) for dic in data] for k in data[0]
                }
            case ():
                self._data = {"": [self.to_child_node(k) for k in data]}
            case None:
                self._data = {}
            case _:
                raise TypeError(data)

    def add_row(
        self,
        row: Sequence[str | None | mk.MkNode] | dict[str, str | None],
    ):
        if len(row) != len(self.columns):
            msg = "Row to add doesnt have same length as header"
            raise ValueError(msg)
        match row:
            case dict():
                for k, v in row.items():
                    self.data[k].append(self.to_child_node(v))
            case _:
                for i, key in enumerate(self.data.keys()):
                    self.data[key].append(self.to_child_node(row[i]))

    def iter_rows(self) -> Iterator[list[mk.MkNode]]:
        data = self.data  # property
        length = min(len(i) for i in data.values())
        for j, _ in enumerate(range(length)):
            yield [data[k][j] for k in data]

    @classmethod
    def for_items(cls, items, columns: dict[str, Callable]):
        ls = [{k: v(item) for k, v in columns.items()} for item in items]
        return cls(ls)

    def width_for_column(self, column: str | int) -> int:
        """Returns the minimum width needed for given column.

        Arguments:
            column: Name or index of the column
        """
        data = self.data  # property
        col_name = list(data.keys())[column] if isinstance(column, int) else column
        col = data[col_name]
        max_len = max((len(str(i).replace("\n", "<br>")) for i in col), default=0)
        return max(len(col_name), max_len)



from __future__ import annotations

from collections.abc import Callable
import dataclasses

from pymdownx import superfences


@dataclasses.dataclass(frozen=True)
class Fences:
    """Class describing a superfence.

    Arguments:
        name: name of the fence
        class_name: CSS class used by the fence
        format_fn: Callable for formatting
    """

    name: str
    class_name: str
    format_fn: str | Callable

    def as_dict(self):
        """Return fence data as a dictionary."""
        return {"name": self.name, "class": self.class_name, "format": self.format_fn}


mermaid_fence = Fences(
    name="mermaid",
    class_name="mermaid",
    format_fn=superfences.fence_code_format,
)

FENCES: dict[str, Fences] = {
    fence.name: fence
    for fence in [
        mermaid_fence,
    ]
}



from __future__ import annotations

from typing import Literal

from mknodes.basenodes import mktable
from mknodes.info import packageinfo, packageregistry
from mknodes.utils import layouts, log, reprhelpers
from mknodes.utils.packagehelpers import Dependency


logger = log.get_logger(__name__)

PackageLayoutStr = Literal["default", "badge"]


class MkDependencyTable(mktable.MkTable):
    """Node for a table showing dependencies for a package."""

    ICON = "material/database"
    STATUS = "new"

    def __init__(
        self,
        package: str | packageinfo.PackageInfo | None = None,
        layout: PackageLayoutStr = "default",
        **kwargs,
    ):
        self.package = package
        match layout:
            case "default":
                self.layouter = layouts.DefaultPackageLayout()
            case "badge":
                self.layouter = layouts.BadgePackageLayout()
            case _:
                raise ValueError(layout)
        super().__init__(**kwargs)

    def __repr__(self):
        return reprhelpers.get_repr(self, package=self.package, _filter_empty=True)

    @property
    def required_packages(self) -> dict[packageinfo.PackageInfo, Dependency] | None:
        match self.package:
            case None:
                return self.ctx.metadata.required_packages
            case str():
                return packageregistry.get_info(self.package).required_packages
            case packageinfo.PackageInfo():
                return self.package.required_packages
            case _:
                return None

    @property
    def data(self):
        if not self.required_packages:
            return {}
        packages = self.required_packages
        if data := [self.layouter.get_row_for(kls) for kls in packages.items()]:
            return {
                k: [self.to_child_node(dic[k]) for dic in data]  # type: ignore[index]
                for k in data[0]
            }
        return {}

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node_1 = MkDependencyTable()
        page += mk.MkReprRawRendered(node_1, header="### From project")
        node_2 = MkDependencyTable("mkdocs")
        page += mk.MkReprRawRendered(node_2, header="### Explicitely defined")
        node_3 = MkDependencyTable(layout="badge")
        page += mk.MkReprRawRendered(node_3, header="### From project")


if __name__ == "__main__":
    table = MkDependencyTable("mknodes")
    logger.warning(table)



from __future__ import annotations

from collections.abc import Mapping
import textwrap

from typing import TYPE_CHECKING, Any

from mknodes.basenodes import mkcontainer, mktext
from mknodes.utils import log, reprhelpers, resources


if TYPE_CHECKING:
    import mknodes as mk

logger = log.get_logger(__name__)


class MkFootNote(mkcontainer.MkContainer):
    """Represents a single footnote. It gets managed by an MkFootNotes node."""

    REQUIRED_EXTENSIONS = [resources.Extension("footnotes")]
    STATUS = "new"

    def __init__(
        self,
        num: int,
        content: str | mk.MkNode,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            num: Footnote index number
            content: Footnote content
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(content=content, **kwargs)
        self.num = num

    def __repr__(self):
        return reprhelpers.get_repr(self, num=self.num, content=self.items)

    def _to_markdown(self) -> str:
        item_str = "\n".join(i.to_markdown() for i in self.items)
        indented = textwrap.indent(item_str, "    ")
        return f"[^{self.num}]:\n{indented}\n"


class MkFootNotes(mkcontainer.MkContainer):
    """Node containing a list of MkFootNotes."""

    items: list[MkFootNote]
    ICON = "octicons/list-ordered-16"
    REQUIRED_EXTENSIONS = [resources.Extension("footnotes")]

    def __init__(
        self,
        footnotes: (
            Mapping[int, str | mk.MkNode] | list[MkFootNote] | list[str] | None
        ) = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            footnotes: Footnotes data (Can be given in different shapes)
            kwargs: Keyword arguments passed to parent
        """
        match footnotes:
            case None:
                items = []
            case list():
                items = [
                    (
                        ann
                        if isinstance(ann, MkFootNote)
                        else MkFootNote(i, ann)  # type: ignore
                    )
                    for i, ann in enumerate(footnotes, start=1)
                ]
            case Mapping():
                items = [
                    MkFootNote(
                        k,
                        content=mktext.MkText(v) if isinstance(v, str) else v,
                    )
                    for k, v in footnotes.items()
                ]
            case _:
                raise TypeError(footnotes)
        super().__init__(content=items, **kwargs)

    def __repr__(self):
        notes: list[mk.MkNode | str] = []
        for item in self.items:
            if len(item.items) == 1 and isinstance(item.items[0], mktext.MkText):
                notes.append(str(item.items[0]))
            elif len(item.items) == 1:
                notes.append(item.items[0])
            else:
                notes.append(item)
        return reprhelpers.get_repr(self, footnotes=notes)

    def __len__(self):
        return len(self.items)

    def __getitem__(self, index: int):
        for node in self.items:
            if node.num == index:
                return node
        raise IndexError(index)

    def __contains__(self, item: int | MkFootNote) -> bool:
        match item:
            case MkFootNote():
                return item in self.items
            case int():
                return any(i.num == item for i in self.items)
            case _:
                raise TypeError(item)

    def _get_item_pos(self, num: int) -> int:
        item = next(i for i in self.items if i.num == num)
        return self.items.index(item)

    def __setitem__(self, index: int, value: mk.MkNode | str):
        match value:
            case MkFootNote():
                node = value
            case _:
                node = MkFootNote(index, content=value, parent=self)
        if index in self:
            pos = self._get_item_pos(index)
            self.items[pos] = node
        else:
            self.items.append(node)

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkFootNotes()
        page += "The MkFootNotes node aggregates footnotes[^1]."
        node[1] = r"Footnotes are numbered, can be set via \__setitem__."
        node[2] = r"They can also get nested[^3]."
        node[3] = mk.MkAdmonition("And they can also contain other Markdown.")
        page += node
        page += mk.MkReprRawRendered(node)

    def _to_markdown(self) -> str:
        if not self.items:
            return ""
        items = sorted(self.items, key=lambda x: x.num)
        return "".join(i.to_markdown() for i in items)


if __name__ == "__main__":
    import mknodes as mk

    # ann = MkFootNote(1, "test")
    # print(ann)
    page = mk.MkPage()
    MkFootNotes.create_example_page(page)
    print(page)



import mknodes as mk


DOC_URL = "https://phil65.github.io/mknodes/"


def create_github_index_md() -> mk.MkPage:
    page = mk.MkPage("Github index")
    page += mk.MkCode.for_object(create_github_index_md)
    page += mk.MkHeader("MkNodes", level=1)
    page += mk.MkHeader("Don't write docs. Code them.", level=4)
    page += mk.MkShields()
    page += mk.MkLink(DOC_URL, "Read the completely coded documentation!")
    page += mk.MkInstallGuide(header="How to install")
    page += mk.MkHeader("All the nodes!")
    page += mk.MkClassDiagram(mk.MkNode, mode="subclasses", direction="LR")
    return page



from __future__ import annotations


def format_js_map(dct: dict) -> str:
    """Return JS map str for given dictionary.

    Arguments:
        dct: Dictionary to dump
    """
    rows = []
    for k, v in dct.items():
        match v:
            case bool():
                rows.append(f"    {k}: {str(v).lower()},")
            case dict():
                rows.append(f"    {k}: {format_js_map(v)},")
            case None:
                rows.append(f"    {k}: null,")
            case _:
                rows.append(f"    {k}: {v!r},")
    row_str = "\n" + "\n".join(rows) + "\n"
    return f"{{{row_str}}}"


if __name__ == "__main__":
    dct = {
        "orientation": "right",
        "scale": 1.5,
        "overflow": True,
        "delay": 0.6,
        "transition": "cubic-bezier(0,0,0,1)",
        "maxTransition": 60,
    }
    result = format_js_map(dct)
    print(result)



from __future__ import annotations

import abc

from collections.abc import Callable
import re
import types
from typing import TYPE_CHECKING

from mknodes.basenodes import mkcontainer, mklink, mklist, mknode
from mknodes.info import linkprovider
from mknodes.templatenodes import mkmetadatabadges
from mknodes.utils import classhelpers, helpers, inspecthelpers, log


if TYPE_CHECKING:
    from mknodes.info import packageinfo
    from mknodes.utils import packagehelpers


logger = log.get_logger(__name__)

MARKER_RE = r'([A-Za-z_]* [>|=|<]* ".*?")'


class Layout(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_row_for(self, obj):
        raise NotImplementedError


class CompactClassLayout(Layout):
    """Table showing info for a list of classes."""

    def __init__(self, link_provider: linkprovider.LinkProvider | None = None):
        self.linkprovider = link_provider or linkprovider.LinkProvider()
        super().__init__()

    def get_columns(self) -> list[str]:
        return ["Class", "Module", "Description"]

    def get_row_for(self, kls: type) -> dict[str, str]:
        return dict(
            Class=self.linkprovider.link_for_klass(kls),
            Module=kls.__module__,
            Description=inspecthelpers.get_doc(kls, only_summary=True),
        )


class ExtendedClassLayout(Layout):
    """Table showing info for a list of classes."""

    def __init__(
        self,
        link_provider: linkprovider.LinkProvider | None = None,
        subclass_predicate: Callable | None = None,
    ):
        self.linkprovider = link_provider or linkprovider.LinkProvider()
        if subclass_predicate is None:

            def always_true(_):
                return True

            self.subclass_predicate = always_true
        else:
            self.subclass_predicate = subclass_predicate
        super().__init__()

    def get_row_for(
        self,
        kls: type,
        shorten_lists_after: int = 10,
    ) -> dict[str, str]:
        """Return a table row for given class.

        Includes columns for child and parent classes including links.
        """
        subclass_links = [
            self.linkprovider.link_for_klass(sub)
            for sub in classhelpers.iter_subclasses(kls, recursive=False)
            if self.subclass_predicate(sub)
        ]
        subclass_str = mklist.MkList(
            subclass_links,
            shorten_after=shorten_lists_after,
        )
        parents = [
            base_kls
            for base_kls in kls.__bases__
            if "<locals>" not in base_kls.__qualname__  # filter locally defined
        ]
        parent_links = [self.linkprovider.link_for_klass(parent) for parent in parents]
        parent_str = mklist.MkList(parent_links, shorten_after=shorten_lists_after)
        desc = inspecthelpers.get_doc(kls, escape=True, only_summary=True)
        link = self.linkprovider.link_for_klass(kls)
        name = helpers.styled(link, size=4, bold=True)
        module = helpers.styled(kls.__module__, size=1, italic=True)
        return dict(
            Name=f"{name}<br>{module}<br>{desc}",
            # Module=kls.__module__,
            Children=subclass_str.to_html(),
            Inherits=parent_str.to_html(),
            # Description=desc,
        )

    def get_columns(self) -> list[str]:
        return ["Name", "Children", "Inherits"]


class ModuleLayout(Layout):
    def __init__(self, link_provider: linkprovider.LinkProvider | None = None):
        self.linkprovider = link_provider or linkprovider.LinkProvider()
        super().__init__()

    def get_row_for(self, module: types.ModuleType) -> dict[str, str]:
        fallback = "*No docstrings defined.*"
        doc = inspecthelpers.get_doc(module, fallback=fallback, only_summary=True)
        return dict(
            Name=self.linkprovider.link_for_module(module),
            DocStrings=doc,
            Members=(
                mklist.MkList(module.__all__, as_links=True, shorten_after=10).to_html()
                if hasattr(module, "__all__")
                else ""
            ),
        )

    def get_columns(self) -> list[str]:
        return ["Name", "DocStrings", "Members"]


class BadgePackageLayout(Layout):
    def get_row_for(
        self,
        dependency: tuple[packageinfo.PackageInfo, packagehelpers.Dependency],
    ) -> dict[str, str | mknode.MkNode]:
        package_info = dependency[0]
        dep_info = dependency[1]
        if url := package_info.homepage:
            link_text = str(mklink.MkLink(url, package_info.name))
        else:
            link_text = f"`{package_info.name}`"
        link = helpers.styled(link_text, size=3, bold=True)
        marker = str(dep_info.marker) if dep_info.marker else ""
        marker_str = re.sub(MARKER_RE, r"`\g<1>`", marker)
        summary = helpers.styled(package_info.metadata["Summary"], italic=True)
        info = mkmetadatabadges.MkMetadataBadges("websites", package=package_info.name)
        info.block_separator = "  "
        container_1 = mkcontainer.MkContainer([link, marker_str], block_separator="\n")
        container_2 = mkcontainer.MkContainer([summary, info])
        return dict(Name=container_1, Summary=container_2)


class DefaultPackageLayout(Layout):
    def get_row_for(
        self,
        dependency: tuple[packageinfo.PackageInfo, packagehelpers.Dependency],
    ) -> dict[str, str | mknode.MkNode]:
        package_info = dependency[0]
        dep_info = dependency[1]
        if url := package_info.homepage:
            node = mklink.MkLink(url, package_info.name)
        else:
            node = f"`{package_info.name}`"
        link = helpers.styled(str(node), size=3, bold=True)
        marker = str(dep_info.marker) if dep_info.marker else ""
        marker_str = re.sub(MARKER_RE, r"`\g<1>`", marker)
        summary = helpers.styled(package_info.metadata["Summary"], italic=True)
        return dict(Name=link, Summary=summary, Markers=marker_str)


if __name__ == "__main__":
    layout = ExtendedClassLayout()
    print(layout.get_row_for(ModuleLayout))



import mknodes as mk

from mknodes import paths
from mknodes.manual import routing
from mknodes.navs import navrouter


DOC_TEXT = """Now lets create the documentation.
This code will show how to build a simple documentation section.
"""

nav = mk.MkNav("MkNavs")


@nav.route.page(is_index=True)
def _(page: mk.MkPage):
    variables = dict(mknode_cls=mk.MkNav)
    page += mk.MkJinjaTemplate("navs/navs_index.jinja", variables=variables)


@nav.route.nav("Populate MkPages from SUMMARY.md")
def _(nav: mk.MkNav):
    folder = paths.TEST_RESOURCES / "nav_tree/"
    summary_file = folder / "SUMMARY.md"
    nav.parse.file(summary_file, hide="toc")
    page = nav.add_page(is_index=True, hide="toc", icon="file")
    text = summary_file.read_text()
    text = text.replace("](", "] (")  ##
    path = paths.TEST_RESOURCES / "nav_tree/"
    variables = dict(path=path, text=text)
    page += mk.MkJinjaTemplate("navs/nav_from_file.jinja", variables=variables)


@nav.route.nav("Populate MkPages from folder")
def _(nav: mk.MkNav):
    """Create a MkNav based on a folder tree containing markup files."""
    folder = paths.TEST_RESOURCES / "nav_tree/test_folder/"
    nav.parse.folder(folder, hide="toc")
    page = nav.add_page(is_index=True, hide="toc", icon="folder")
    variables = dict(folder=folder)
    page += mk.MkJinjaTemplate("navs/nav_from_folder.jinja", variables=variables)


@nav.route.nav("Routing")
def _(nav: mk.MkNav):
    page = routing.nav.add_page(is_index=True, icon="material/call-split", hide="toc")
    page += mk.MkCode.for_file(routing.__file__, header="Code for this section")
    page += mk.MkDocStrings(navrouter.NavRouter, header="MkNav.route Docstrings")
    return routing.nav


# @nav.route.nav("MkDefaultWebsite")
def _(nav: mk.MkNav):
    proj = mk.Project.for_path("https://github.com/mkdocstrings/mkdocstrings.git")
    nav += mk.MkDefaultWebsite(section="MkDocStrings", context=proj.context)


@nav.route.nav("The MkDoc class")
def _(nav: mk.MkNav):
    page = nav.add_page(is_index=True, hide="toc", icon="api")
    page += mk.MkAdmonition(DOC_TEXT, typ="tip")
    template = "classpage_custom.jinja"
    nav.add_doc(module=mk, class_page=template, recursive=True)


@nav.route.page("Context propagation")
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("context_propagation.jinja")



from __future__ import annotations

from typing import Any, Self

from mknodes.basenodes import mknode
from mknodes.utils import helpers, log, pathhelpers, reprhelpers


logger = log.get_logger(__name__)

RESPONSE_CODE_OK = 200
EXAMPLE_URL = "https://raw.githubusercontent.com/phil65/mknodes/main/README.md"


class MkText(mknode.MkNode):
    """Class for any Markup text.

    All classes inheriting from MkNode can get converted to this Type.
    """

    ICON = "material/text"

    def __init__(
        self,
        text: str | mknode.MkNode | None = "",
        is_jinja_expression: bool = False,
        *,
        header: str = "",
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            text: Markup text
            is_jinja_expression: Whether text is a jinja expression
            header: Section header
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(header=header, **kwargs)
        self._text = str(text or "")
        self.is_jinja_expression = is_jinja_expression

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            text=self.text,
            is_jinja_expression=self.is_jinja_expression,
            _filter_false=True,
        )

    def __getitem__(self, section_name: str) -> Self | None:
        markdown = self._to_markdown()
        section_text = helpers.extract_header_section(markdown, section_name)
        return None if section_text is None else type(self)(section_text)

    @property
    def text(self) -> str:
        if not self.is_jinja_expression:
            return self._text
        return self.env.render_string(f"{{{{ {self._text} }}}}")

    @text.setter
    def text(self, value):
        self._text = value

    def _to_markdown(self) -> str:
        return self.text

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkText("This is the most basic node. It contains `markdown` text")
        page += mk.MkReprRawRendered(node, header="### Regular")
        if from_url := MkText.from_url(EXAMPLE_URL):
            page += mk.MkReprRawRendered(from_url, header="### From URL")

    @classmethod
    def from_url(cls, url: str) -> Self | None:
        """Build a MkText node on a remote markup file.

        If the URL contains a "#" (http://.../markdown.md#section),
        it will try to extract the given section.

        All fsspec protocols are supported as URL.

        Arguments:
            url: URL to get markdown from.
        """
        url, *section = url.split("#")
        text = pathhelpers.load_file_cached(url)
        if section:
            text = helpers.extract_header_section(text, section[0])
        return cls(text) if text is not None else None


if __name__ == "__main__":
    node = MkText("log", is_jinja_expression=True)
    print(node)



from __future__ import annotations

import itertools
import textwrap

from typing import Any, Literal

from pymdownx import superfences

from mknodes.basenodes import mkcode
from mknodes.utils import helpers, reprhelpers, resources


GraphTypeStr = Literal["flow", "sequence", "state"]

config = {
    "custom_fences": [
        {"name": "mermaid", "class": "mermaid", "format": superfences.fence_code_format},
    ],
}


class MkDiagram(mkcode.MkCode):
    """Class representing a mermaid diagram.

    MkDiagrams can show directed acyclic graphs and allows to manually
    create diagrams.
    """

    ICON = "material/graph-outline"
    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.superfences", **config)]

    TYPE_MAP = dict(
        flow="graph",
        sequence="sequenceDiagram",
        state="stateDiagram-v2",
    )
    ORIENTATION = dict(
        default="",
        left_right="LR",
        top_down="TD",
        right_left="RL",
        down_top="DT",
    )

    def __init__(
        self,
        items: list | None = None,
        connections: list[tuple] | None = None,
        *,
        graph_type: GraphTypeStr = "flow",
        direction: Literal["TD", "DT", "LR", "RL"] = "TD",
        attributes: dict[str, str] | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            graph_type: Type of the graph
            items: items which should be part of the diagram
            connections: tuples indicating the connections of the items
            direction: diagram direction
            attributes: Optional attributes for the items
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(language="mermaid", **kwargs)
        self._graph_type = graph_type
        self._direction = direction
        # Preserve order. Useful if only items are passed, order is important then.
        self.names = sorted(set(items or []), key=items.index) if items else []
        self.connections = set(connections or [])
        self.attributes = attributes or {}

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            graph_type=self.graph_type,
            items=self.names,
            connections=self.connections,
            direction=self.direction,
        )

    @property
    def graph_type(self):
        """The type of the graph (usually flow)."""
        return (
            self._graph_type
            if self._graph_type not in self.TYPE_MAP
            else self.TYPE_MAP[self._graph_type]
        )

    @property
    def direction(self):
        """The graph direction."""
        return (
            self._direction
            if self._direction not in self.ORIENTATION
            else self.ORIENTATION[self._direction]
        )

    @property
    def text(self):
        """MkCode override."""
        return f"{self.graph_type} {self.direction}\n{self.mermaid_code}"

    @property
    def mermaid_code(self) -> str:
        """Return code block, excluding fences and (graph type direction) line.

        Can be overriden by subclasses.
        """
        lines = list(self.names)
        if not self.connections:
            lines = [f'{helpers.get_hash(i)}["{i}"]' for i in lines]
            for prev, nxt in itertools.pairwise(self.names):
                lines.append(f"{helpers.get_hash(prev)} --> {helpers.get_hash(nxt)}")
            return textwrap.indent("\n".join(lines), "  ")
        for connection in self.connections:
            if len(connection) == 2:  # noqa: PLR2004
                source, target = connection
                lines.append(f"{source} --> {target}")
            elif len(connection) == 3:  # noqa: PLR2004
                source, target, mark = connection
                lines.append(f"{source} --> |{mark}| {target}")
            else:
                msg = f"Tuple with wrong length: {connection}"
                raise TypeError(msg)
        return textwrap.indent("\n".join(lines), "  ")

    @property
    def fence_title(self):
        """MkCode override."""
        return "mermaid"

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        page += "MkDiagrams can be used to create Mermaid diagrams manually."
        diagram = MkDiagram(items=["1", "2", "3"], connections=[("1", "2"), ("2", "3")])
        page += mk.MkReprRawRendered(diagram, header="### Regular")
        diagram = MkDiagram(
            items=["1", "2", "3"],
            connections=[("1", "2"), ("1", "3", "comment")],
            direction="LR",
        )
        page += mk.MkReprRawRendered(diagram, header="### Direction")


if __name__ == "__main__":
    diagram = MkDiagram(["a", "b", "c", "d"])
    print(diagram)



"""MkNodes project."""

from __future__ import annotations

from collections.abc import Callable
import os
import pathlib

from typing import Generic, TypeVar

from mknodes import paths
from mknodes.info import contexts, folderinfo, linkprovider, reporegistry
from mknodes.jinja import environment
from mknodes.navs import mknav
from mknodes.theme import theme as theme_
from mknodes.utils import classhelpers, log, reprhelpers


logger = log.get_logger(__name__)


T = TypeVar("T", bound=theme_.Theme)


class Project(Generic[T]):
    """MkNodes Project."""

    def __init__(
        self,
        theme: T,
        repo: str | os.PathLike | None | folderinfo.FolderInfo = None,
        build_fn: str | Callable = paths.DEFAULT_BUILD_FN,
        base_url: str = "",
        use_directory_urls: bool = True,
        clone_depth: int = 100,
        jinja_extensions: list[str] | None = None,
    ):
        """The main project to create a website.

        Arguments:
            theme: The theme to use
            repo: Path to the git repository
            build_fn: Callable to create the website with
            base_url: Base url of the website
            use_directory_urls: Whether urls are in directory-style
            clone_depth: Amount of commits to clone in case repository is remote.
            jinja_extensions: Optional additional jinja extensions to load.
        """
        self.linkprovider = linkprovider.LinkProvider(
            base_url=base_url,
            use_directory_urls=use_directory_urls,
            include_stdlib=True,
        )
        self.build_fn = classhelpers.to_callable(build_fn)
        self.theme: T = theme
        git_repo = reporegistry.get_repo(str(repo or "."), clone_depth=clone_depth)
        self.folderinfo = folderinfo.FolderInfo(git_repo.working_dir)
        self.context = contexts.ProjectContext(
            metadata=self.folderinfo.context,
            git=self.folderinfo.git.context,
            # github=self.folderinfo.github.context,
            theme=self.theme.context,
            links=self.linkprovider,
            env=environment.Environment(load_templates=True),
        )
        self._root = mknav.MkNav(context=self.context)

    def __repr__(self):
        return reprhelpers.get_repr(self, repo_path=str(self.folderinfo.path))

    @property
    def root(self) -> mknav.MkNav:
        return self._root

    @root.setter
    def root(self, nav: mknav.MkNav):
        self._root = nav
        nav._ctx = self.context

    @classmethod
    def for_path(cls, path: str) -> Project:
        instance = cls(
            base_url="",
            use_directory_urls=True,
            theme=theme_.Theme.get_theme("material", data={}),
            repo=path,
        )
        logger.debug("Building page...")
        instance.build_fn(project=instance)
        logger.debug("Finished building page.")
        if not instance._root:
            msg = "No root for project created."
            raise RuntimeError(msg)
        paths = [
            pathlib.Path(node.resolved_file_path).stem
            for _level, node in instance._root.iter_nodes()
            if hasattr(node, "resolved_file_path")
        ]
        instance.linkprovider.set_excludes(paths)
        return instance

    def set_root(self, nav: mknav.MkNav):
        """Set the root MkNav."""
        self._root = nav
        nav._ctx = self.context

    def get_root(self) -> mknav.MkNav:
        """Return the root MkNav.

        This MkNav should get populated in order to build
        the website.

        Arguments:
            kwargs: Keyword arguments passed to MkNav constructor.
        """
        return self._root


if __name__ == "__main__":
    import mknodes as mk

    from mknodes.manual import root

    theme = mk.MaterialTheme()
    project = Project(theme=theme)
    log.basic()
    root.build(project)



import typing

import mknodes as mk

from mknodes.data import datatypes
from mknodes.manual.a_quick_tour import a_quick_tour


nav = mk.MkNav("Get started")


@nav.route.page("Welcome to MkNodes", hide="toc", is_homepage=True)
def _(page: mk.MkPage):
    page += mk.MkShields(["version", "status", "codecov"])
    fn_code = mk.MkCode.for_file(__file__)
    node = mk.MkAdmonition(content=fn_code, title="")
    for i in typing.get_args(datatypes.AdmonitionTypeStr):
        node = mk.MkAdmonition(content=node, typ=i, title="")
    page += node


@nav.route.page("Installation", hide="toc", icon="grommet-icons:install")
def _(page: mk.MkPage):
    page += mk.MkInstallGuide()


@nav.route.page("Why should I use MkNodes?", hide="toc", icon="ri:question-line")
def _(page: mk.MkPage):
    page += mk.MkJinjaTemplate("why_mknodes.jinja")


@nav.route.page("A quick node tour", hide="toc", icon="ic:outline-tour")
def _(page: mk.MkPage):
    a_quick_tour(page)


@nav.route.page("Changelog", icon="format-list-group")
def _(page: mk.MkPage):
    page += mk.MkChangelog()



from __future__ import annotations

import mknodes as mk


def build(project):
    sub_nav = mk.MkNav("Sub nav")
    sub_nav.page_template.announcement_bar = "Hello"
    sub_nav += mk.MkPage("Test page")
    project.root += sub_nav


# def test_templates():
#     theme = mk.MaterialTheme()
#     project = mk.Project(theme=theme, repo=".", build_fn=build)
#     assert project._root
#     project._root.get_resources()
#     assert len([i for i in resources.templates if i]) == 1



from __future__ import annotations

from collections.abc import Hashable
import textwrap

from mknodes.utils import log


logger = log.get_logger(__name__)


class Connector:
    def __init__(self, objects, max_depth: int | None = None):
        if not isinstance(objects, list | tuple):
            objects = [objects]
        self.item_dict: dict[Hashable, str] = {}
        self.connections: list[tuple[Hashable, Hashable]] = []
        self.max_depth = max_depth
        self._connect(objects)

    def _connect(self, objects):
        def add_connections(item, depth: int = 0):
            identifier = self.get_id(item)
            if identifier not in self.item_dict:
                # if item.__module__.startswith(base_module):
                self.item_dict[identifier] = self.get_title(item)
                if self.max_depth and self.max_depth <= depth:
                    return
                for base in self.get_children(item):
                    self.connections.append((self.get_id(base), identifier))
                    add_connections(base, depth + 1)

        for obj in objects:
            add_connections(obj)

    @property
    def items(self):
        return list(self.item_dict.keys())

    @property
    def titles(self):
        return list(self.item_dict.values())

    def get_children(self, item) -> list | tuple:
        """This should return a list of children for the tree node."""
        return NotImplemented

    def get_id(self, item):
        """This needs to return a unique identifier for an item."""
        return item

    def get_attributes(self, item):
        return None

    def get_title(self, item) -> str:
        """This can be overridden for a nicer label."""
        return str(self.get_id(item))

    def get_graph_connection_text(self):
        lines = [
            f'{identifier}["{title}"]'
            for identifier, title in zip(self.items, self.titles)
        ]
        lines += [f"{a} --> {b}" for a, b in self.connections]
        return textwrap.indent("\n".join(lines), "  ")


if __name__ == "__main__":

    class Test(Connector):
        def get_children(self, item):
            return item.__bases__

    test = Test(Connector).get_graph_connection_text()
    print(test)



"""Link replacer."""

from __future__ import annotations

import collections
import re
import urllib.parse

from mknodes.utils import log, pathhelpers


logger = log.get_logger(__name__)


# For Regex, match groups are:
#       0: Whole markdown link e.g. [Alt-text](url)
#       1: Alt text
#       2: Full URL e.g. url + hash anchor
#       3: Filename e.g. filename.md
#       4: File extension e.g. .md, .png, etc.
#       5. hash anchor e.g. #my-sub-heading-link
AUTOLINK_RE = r"\[([^\]]+)\]\((([^)/]+\.(md|png|jpg))(#.*)*)\)"


class LinkReplacer:
    def __init__(self):
        self.mapping = collections.defaultdict(list)
        self.page_url = ""

    def __call__(self, match) -> str:
        filename = urllib.parse.unquote(match.group(3).strip())
        if filename not in self.mapping:
            return f"`{match.group(1)}`"
        filenames = self.mapping[filename]
        if len(filenames) > 1:
            text = "%s: %s has %s targets"
            logger.debug(text, self.page_url, len(filenames), match.group(3))
        new_link = pathhelpers.relative_url(self.page_url, filenames[0])
        return match.group(0).replace(match.group(2), new_link)
        # new_text = new_text.replace("\\", "/")
        # text = "LinkReplacer: : %s -> %s"
        # logger.debug(text, match.group(3), new_text)

    def replace(self, markdown: str, uri: str) -> str:
        if uri.endswith("SUMMARY.md"):
            return markdown
        self.page_url = uri
        return re.sub(AUTOLINK_RE, self, markdown)



from __future__ import annotations

import mknodes as mk


EXPECTED = """## test

  * a
  * b
"""


def test_list():
    ls = mk.MkList()
    assert not str(ls)


def test_markdown():
    ls = mk.MkList(["a", "b"], header="test")
    assert str(ls) == EXPECTED



from __future__ import annotations

import mknodes as mk


EXPECTED = """## Header

1.  abcde
    fghi
2.  abcde
    fghi
3.  abcde
    fghi
4.  abcde
    fghi
5.  abcde
    fghi
6.  abcde
    fghi
7.  abcde
    fghi
8.  abcde
    fghi
9.  abcde
    fghi
10. abcde
    fghi
"""

EXPECTED_SORTED = """1.  1
2.  2
"""


def test_empty():
    annotation = mk.MkAnnotations()
    assert not str(annotation)


def test_if_annotations_get_sorted():
    node = mk.MkAnnotations()
    node[2] = "2"
    node[1] = "1"
    assert str(node) == EXPECTED_SORTED


def test_markdown():
    annotation = mk.MkAnnotations(["abcde\nfghi"] * 10, header="Header")
    assert str(annotation) == EXPECTED


def test_constructors():
    annotation_1 = mk.MkAnnotations(["abc", "def"])
    anns = {1: "abc", 2: "def"}
    annotation_2 = mk.MkAnnotations(anns)
    assert str(annotation_1) == str(annotation_2)


def test_mapping_interface():
    ann = mk.MkAnnotations()
    ann[1] = "test"
    assert str(ann[1]) == "1.  test\n"


def test_code_annotations_inside_admonition():
    expected = """!!! info
    <div class="annotate" markdown>
    ``` python
    test # (1)
    ```
    </div>

    1.  Some annotation

"""
    code = mk.MkCode("test # (1)", language="python")
    code.annotations[1] = "Some annotation"
    admonition = mk.MkAdmonition(code)
    assert str(admonition) == expected



from __future__ import annotations

import ast
import os
import pathlib

from posixpath import join as urljoin
import re
from typing import TYPE_CHECKING, Any
from urllib import parse

from mknodes.basenodes import mkcode
from mknodes.navs import mknav
from mknodes.pages import mkpage
from mknodes.utils import classhelpers, helpers, log


if TYPE_CHECKING:
    import mknodes as mk


logger = log.get_logger(__name__)

# for SUMMARY.md parsing
SECTION_AND_FILE_RE = r"^\* \[(.*)\]\((.*)\)"
SECTION_AND_FOLDER_RE = r"^\* \[(.*)\]\((.*)\/\)"
SECTION_RE = r"^\* (.*)"

# for ->MkNode parsing
ARGS_KWARGS_RE = r".*\((.*)\)"  # get brace content


def str2page(
    path: str | os.PathLike,
    name: str | None = None,
    **kwargs,
) -> mk.MkPage:
    """Parse given path, check for our -> syntax, and return a MkPage.

    If -> is detected, return a MkPage containing given MkNode. Otherwise
    open / download the markdown file from given path and put that into an MkPage.

    Arguments:
        path: Path to build a MkPage for (Either path / URL or "-> MkNode")
        name: Name for given MkPage
        kwargs: Additional metadata for MkPage
    """
    import mknodes as mk

    path = os.fspath(path)
    logger.debug("Adding file %r", path)
    node_cls_name = path.removeprefix("->").strip().split("(")[0]
    if path.startswith("->") and node_cls_name in mk.__all__:
        node_cls = getattr(mk, node_cls_name)
        page = mk.MkPage(name, **kwargs)
        if match := re.match(ARGS_KWARGS_RE, path):
            parts = match[1].split(",")
            args = [ast.literal_eval(i.strip()) for i in parts if "=" not in i]
            kwargs_iter = (i.strip().split("=", maxsplit=1) for i in parts if "=" in i)
            kwargs = {i[0]: ast.literal_eval(i[1]) for i in kwargs_iter}
            msg = "Parsed: Node: %s, args: %s, kwargs: %s"
            logger.debug(msg, node_cls.__name__, args, kwargs)
            page += node_cls(*args, **kwargs)
        else:
            page += node_cls()
        return page
    return mk.MkPage.from_file(path, title=name, **kwargs)


def from_list(
    ls: list,
    nav: mk.MkNav,
    base_path: str = "",
):
    """Parse given list recursively and add found content to given MkNav.

    Arguments:
        ls: List to parse
        nav: MkNav to attach found stuff to
        base_path: All found paths will be relative to this one if given.
    """
    for item in ls:
        match item:
            case dict():
                name, val = next(iter(item.items()))
                match val:
                    case dict():
                        logger.debug("Adding nav %r", name)
                        from_dict(val, nav.add_nav(name), base_path=base_path)
                    case str():
                        path = urljoin(base_path, val)
                        nav += str2page(path=path, name=name)
                    case list():
                        logger.debug("Adding nav %r", name)
                        if helpers.is_url(name):
                            path = parse.urlsplit(name).path
                            name = pathlib.Path(path).name
                        from_list(val, nav.add_nav(name), base_path=base_path)
            case str():
                path = urljoin(base_path, item)
                page = mkpage.MkPage.from_file(path)
                logger.debug("Adding page %s", path)
                nav += page


def from_dict(
    dct: dict[str, str | list | dict],
    nav: mk.MkNav,
    base_path: str = "",
):
    """Parse given dict recursively and add found content to given MkNav.

    Arguments:
        dct: Dictionary to parse
        nav: MkNav to attach found stuff to
        base_path: All found paths will be relative to this one if given.
    """
    base_path = base_path.replace("\\", "/")
    for k, v in dct.items():
        match v:
            case str():
                path = urljoin(base_path, v)
                nav += str2page(path=path, name=k)
            case dict():
                logger.debug("Adding nav %r", k)
                from_dict(v, nav.add_nav(k), base_path=base_path)
            case list():
                logger.debug("Adding nav %r", k)
                from_list(v, nav.add_nav(k), base_path=base_path)


class NavParser:
    """Class used for constructing MkNavs."""

    def __init__(self, nav: mk.MkNav):
        """Constructor.

        Arguments:
            nav: MkNav to use for routing
        """
        self._nav = nav

    def json(
        self,
        obj: list | dict[str, str | list | dict],
    ):
        """Parse given list or dict and attach it to given MkNav.

        If no Nav is given, create a new one
        Arguments:
            obj: Dictionary / List to parse
        """
        match obj:
            case dict():
                from_dict(obj, self._nav)
            case list():
                from_list(obj, self._nav)

    def file(
        self,
        path: str | os.PathLike,
        **kwargs: Any,
    ) -> mk.MkNav:
        """Load an existing SUMMARY.md style file.

        For each indentation level in SUMMARY.md, a new sub-nav is created.

        Should support all SUMMARY.md options except wildcards.

        Arguments:
            path: Path to the file
            kwargs: Keyword arguments passed to the pages to create.
                    Can be used to hide the TOC for all pages for example.
        """
        path = pathlib.Path(path)
        if path.is_absolute():
            path = os.path.relpath(path, pathlib.Path().absolute())
            path = pathlib.Path(path)
        return self.text(
            path.read_text(),
            path=path,
            **kwargs,
        )

    def text(
        self,
        text: str,
        path: pathlib.Path,
        **kwargs: Any,
    ) -> mk.MkNav:
        """Create a nav based on a SUMMARY.md-style list, given as text.

        For each indentation level, a new sub-nav is created.

        Should support all SUMMARY.md options except wildcards.

        Arguments:
            text: Text to parse
            path: path of the file containing the text.
            kwargs: Keyword arguments passed to the pages to create.
                    Can be used to hide the TOC for all pages for example.
        """
        lines = text.splitlines()
        for i, line in enumerate(lines):
            # * [Example](example_folder/)

            if m := re.match(SECTION_AND_FOLDER_RE, line):
                folder_path = path.parent / m[2]
                if (file_path := (folder_path / "SUMMARY.md")).exists():
                    subnav = mknav.MkNav(m[1], parent=self._nav)
                    subnav.parse.file(file_path, **kwargs)
                elif (file_path := (folder_path / "index.py")).exists():
                    mod = classhelpers.import_file(file_path)
                    subnav = mod.nav
                    subnav.parent = self._nav
                else:
                    msg = "No SUMMARY.md / index.py found."
                    raise RuntimeError(msg)
                self._nav[m[1]] = subnav
                logger.debug("Created subsection %s from %s", m[1], file_path)

            # * [Example](example_folder/sub_1.md)

            elif m := re.match(SECTION_AND_FILE_RE, line):
                # if following section indented, it is a nav with an index page:
                if unindented := helpers.get_indented_lines(lines[i + 1 :]):
                    subnav = mknav.MkNav(m[1], parent=self._nav)
                    text = "\n".join(unindented)
                    subnav.parse.text(text, path=path, **kwargs)
                    page = subnav.add_page(is_index=True, **kwargs)
                    page += pathlib.Path(m[2]).read_text()
                    msg = "Created subsection %s and loaded index page %s"
                    logger.debug(msg, m[1], m[2])
                    self._nav += subnav

                # if not, add as regular page:
                else:
                    p = m[2] if m[2].startswith("->") else path.parent / m[2]
                    page = str2page(path=p, name=m[1], parent=self._nav, **kwargs)
                    self._nav[m[1]] = page
                    logger.debug("Created page %s from %s", m[1], m[2])

            # * Section/

            elif m := re.match(SECTION_RE, line):
                unindented = helpers.get_indented_lines(lines[i + 1 :]) or []
                logger.debug("Created subsection %s from text", m[1])
                subnav = mknav.MkNav(m[1], parent=self._nav)
                text = "\n".join(unindented)
                subnav.parse.text(text, path=path, **kwargs)
                self._nav[m[1]] = subnav
        return self._nav

    def folder(
        self,
        folder: str | os.PathLike,
        *,
        recursive: bool = True,
        **kwargs: Any,
    ) -> mk.MkNav:
        """Load a MkNav tree from Folder.

        SUMMARY.mds are ignored.
        index.md files become index pages.

        To override the default behavior of using filenames as menu titles,
        the pages can set a title by using page metadata.

        Arguments:
            folder: Folder to load .md files from
            recursive: Whether all .md files should be included recursively.
            kwargs: Keyword arguments passed to the pages to create.
                    Can be used to hide the TOC for all pages for example.
        """
        import mknodes as mk

        folder = pathlib.Path(folder)
        nav = self._nav
        for path in folder.iterdir():
            is_hidden = path.name.startswith(("_", "."))
            if recursive and path.is_dir() and not is_hidden and any(path.iterdir()):
                path = folder / path.parts[-1]
                subnav = mk.MkNav(path.name)
                subnav.parse.folder(folder=path, **kwargs)
                nav += subnav
                logger.debug("Loaded subnav from from %s", path)
            elif path.name == "index.md":
                logger.debug("Loaded index page from %s", path)
                text = path.read_text(encoding="utf-8")
                nav.index_page = mk.MkPage(path=path.name, content=text, **kwargs)
                nav.index_title = nav.section or "Home"
            elif path.suffix in [".md", ".html"] and path.name != "SUMMARY.md":
                text = path.read_text(encoding="utf-8")
                rel_path = path.relative_to(folder)
                nav += mk.MkPage(path=rel_path, content=text, **kwargs)
                logger.debug("Loaded page from from %s", path)
        return nav

    def module(
        self,
        module: str | os.PathLike,
        *,
        recursive: bool = True,
        **kwargs: Any,
    ) -> mk.MkNav:
        """Load a MkNav tree from a Module.

        Will add a page for each module showing the code in a code box.

        Arguments:
            module: Module to load code files for
            recursive: Whether all .md files should be included recursively.
            kwargs: Keyword arguments passed to the pages to create.
                    Can be used to hide the TOC for all pages for example.
        """
        # if isinstance(module, types.ModuleType):
        #     module = inspecthelpers.get_file(module)
        for path in pathlib.Path(module).iterdir():
            is_hidden = path.name.startswith(("_", "."))
            if recursive and path.is_dir() and not is_hidden and any(path.iterdir()):
                subnav = mknav.MkNav(path.name)
                subnav.parse.folder(folder=path, **kwargs)
                self._nav += subnav
                logger.debug("Loaded subnav from from %s", path)
            elif path.suffix in [".py"]:
                page = mkpage.MkPage(path=path.name, title=path.name, **kwargs)
                content = path.read_text(encoding="utf-8")
                page += mkcode.MkCode(content, language="py", linenums=1)
                self._nav += page
                logger.debug("Loaded page from from %s", path)
        return self._nav


if __name__ == "__main__":
    log.basic()
    nav = mknav.MkNav()
    nav.parse.module("mknodes/manual/")
    print(nav)



from __future__ import annotations

from typing import Any

from mknodes.basenodes import mkcontainer, mknode
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkTask(mkcontainer.MkContainer):
    """Node for a single task listitem."""

    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.tasklist")]
    ICON = "material/library"

    def __init__(
        self,
        value: bool = False,
        content: list | None | str | mknode.MkNode = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            value: Setting value
            content: Markdown content for this block
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(content=content, **kwargs)
        self.value = value

    def __repr__(self):
        return reprhelpers.get_repr(self, value=self.value, content=self.items)

    def _to_markdown(self) -> str:
        lines = super()._to_markdown().split("\n")
        val = "x" if self.value else " "
        result = [f"- [{val}] {lines[0]}"]
        result.extend(f"      {i}" for i in lines[1:])
        return "\n".join(result) + "\n"


class MkTaskList(mkcontainer.MkContainer):
    """Node for task lists."""

    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.tasklist")]
    ICON = "material/library"

    def __init__(
        self,
        content: list | None | str | mknode.MkNode = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            content: Data show for the table
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(content=content, block_separator="", **kwargs)

    def add_item(self, content, value):
        item = MkTask(value, content)
        self.append(item)

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkTaskList()
        node.add_item("True!", True)
        node.add_item("False!", False)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    ls = MkTaskList()
    ls.add_item("True", True)
    ls.add_item("False", False)
    print(ls)



from __future__ import annotations

import functools
import html

from typing import Any, Literal

from mknodes.basenodes import mkimage
from mknodes.utils import helpers, log, reprhelpers, resources


logger = log.get_logger(__name__)


StyleStr = Literal["default", "gitlab-scoped"]


@functools.cache
def get_badge(
    label: str = "",
    value: str = "",
    font_size: int | None = None,
    font_name: str | None = None,
    num_padding_chars: int | None = None,
    badge_color: str | None = None,
    text_color: str | None = None,
    use_gitlab_style: bool = False,
) -> str:
    import anybadge

    badge = anybadge.Badge(
        label=html.escape(label),
        value=html.escape(value),
        font_size=font_size,
        font_name=font_name,
        num_padding_chars=num_padding_chars,
        default_color=badge_color,
        text_color=text_color,
        style="gitlab-scoped" if use_gitlab_style else "default",
    )
    return badge.badge_svg_text


class MkBadge(mkimage.MkImage):
    """Node for a locally-created badge (based on "anybadge").

    The node creates a badge svg, appends it to the virtual files, and
    shows it as an image.
    """

    ICON = "simple/shieldsdotio"
    REQUIRED_PACKAGES = [resources.Package("anybadge")]

    def __init__(
        self,
        label: str | tuple[str, str],
        value: str | None = None,
        *,
        font_size: Literal[10, 11, 12] | None = None,
        font_name: str | None = None,
        num_padding_chars: int | None = None,
        badge_color: str | None = None,
        text_color: str | None = None,
        use_gitlab_style: bool = False,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            label: Left part of the badge. If given a tuple, use 2nd item as value.
            value: Right part of the badge
            font_size: Size of font to use
            font_name: Name of font to use
            num_padding_chars: Number of chars to use for padding
            badge_color: Badge color
            text_color: Badge color
            use_gitlab_style: Use Gitlab-scope style
            kwargs: Keyword arguments passed to parent
        """
        super().__init__("", **kwargs)
        if isinstance(label, tuple):
            self.label, self.value = label
        else:
            self.label = label
            self.value = value
        self.font_size = font_size
        self.font_name = font_name
        self.num_padding_chars = num_padding_chars
        self._badge_color = badge_color
        self._text_color = text_color
        self.use_gitlab_style = use_gitlab_style

    @property
    def badge_color(self) -> str | None:
        if isinstance(self._badge_color, str):
            return self._badge_color
        return self.ctx.theme.primary_color

    def _to_markdown(self):
        data = self.data.replace('<?xml version="1.0" encoding="UTF-8"?>', "")
        content = data.replace("\n", "")
        inner = f"<a href={self.url!r}>{content}</a>" if self.url else content
        return f"<body>{inner}</body>"

    @property
    def text_color(self) -> str | None:
        if isinstance(self._text_color, str):
            return self._text_color
        color = self.ctx.theme.text_color
        return f"{color},#fff" if self.use_gitlab_style else f"#fff,{color}"

    @property
    def data(self) -> str:
        return get_badge(
            label=self.label,
            value=self.value,
            font_size=self.font_size,
            font_name=self.font_name,
            num_padding_chars=self.num_padding_chars,
            badge_color=self.badge_color,
            text_color=self.text_color,
            use_gitlab_style=self.use_gitlab_style,
        )

    @data.setter
    def data(self, value):
        pass

    @property
    def path(self) -> str:
        hashed = helpers.get_hash(repr(self))
        unique = f"{self.label}_{self.value}_{hashed}.svg"
        return helpers.slugify(unique)

    @path.setter
    def path(self, value):
        pass

    def __repr__(self):
        return reprhelpers.get_repr(
            self,
            label=self.label,
            value=self.value,
            font_size=self.font_size,
            font_name=self.font_name,
            badge_color=self.badge_color,
            text_color=self.text_color,
            num_padding_chars=self.num_padding_chars,
            use_gitlab_style=self.use_gitlab_style,
            _filter_empty=True,
            _filter_false=True,
        )

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkBadge(label="Some", value="Badge")
        page += mk.MkReprRawRendered(node)
        node = MkBadge(label="Some", value="Badge", font_size=12)
        page += mk.MkReprRawRendered(node)
        node = MkBadge(label="Some", value="Badge", num_padding_chars=5)
        page += mk.MkReprRawRendered(node)
        node = MkBadge(label="Some", value="Badge", badge_color="teal")
        page += mk.MkReprRawRendered(node)
        node = MkBadge(label="Some", value="Badge", use_gitlab_style=True)
        page += mk.MkReprRawRendered(node)


if __name__ == "__main__":
    img = MkBadge("Left", "right")
    print(img.data)



from __future__ import annotations

import dataclasses
import datetime
import json
import os

import dateutil.parser

from mknodes.theme import mkblog
from mknodes.utils import downloadhelpers


@dataclasses.dataclass(frozen=True)
class Commit:
    sha: str
    author_name: str
    author_email: str
    author_date: datetime.datetime
    author_avatar: str
    author_url: str
    author_login: str
    committer_name: str
    committer_email: str
    committer_date: datetime.datetime
    committer_avatar: str
    committer_url: str
    committer_login: str
    html_url: str
    message: str
    verified: str
    verified_reason: str


def get_latest_commits(owner: str, repo: str, page: int = 1) -> list[Commit]:
    """Return latest commits from given repository.

    Arguments:
        owner: Repository owner
        repo: Repository name
        page: page to get. Each page contains max 100 commits.
    """
    url = f"https://api.github.com/repos/{owner}/{repo}/commits?per_page=100&page={page}"
    response = downloadhelpers.download(url)
    commits = json.loads(response.decode())
    return [
        Commit(
            sha=dct["commit"]["tree"]["sha"],
            author_name=dct["commit"]["author"]["name"],
            author_email=dct["commit"]["author"]["email"],
            author_date=dateutil.parser.parse(dct["commit"]["author"]["date"]),
            author_avatar=dct["author"]["avatar_url"],
            author_login=dct["author"]["login"],
            author_url=dct["author"]["html_url"],
            committer_name=dct["commit"]["committer"]["name"],
            committer_email=dct["commit"]["committer"]["email"],
            committer_date=dateutil.parser.parse(dct["commit"]["committer"]["date"]),
            committer_avatar=dct["committer"]["avatar_url"],
            committer_login=dct["committer"]["login"],
            committer_url=dct["committer"]["html_url"],
            html_url=dct["html_url"],
            message=dct["commit"]["message"],
            verified=dct["commit"]["verification"]["verified"],
            verified_reason=dct["commit"]["verification"]["reason"],
        )
        for dct in commits
    ]


class MkGitBlog(mkblog.MkBlog):
    """The MkDocs-Material plugin-blog, misused as a Git log."""

    def __init__(self, org: str, repo: str, posts_dir: str | os.PathLike, **kwargs):
        super().__init__(**kwargs)
        self.commits: list[mkblog.MkBlogPost] = []
        self.posts_dir = posts_dir
        self.org = org
        self.repo = repo

    def add_commits(self):
        """Fetch commits and add them to the blog."""
        commits = get_latest_commits(self.org, self.repo)
        for c in commits:
            if c.author_login not in self.authors:
                self.authors[c.author_login] = mkblog.Author(
                    name=c.author_login,
                    description=c.author_email,
                    avatar=c.author_avatar,
                )
            if c.committer_login not in self.authors:
                self.authors[c.committer_login] = mkblog.Author(
                    name=c.committer_login,
                    description=c.committer_email,
                    avatar=c.committer_avatar,
                )
            self.add_post(
                title=c.message,
                date=c.author_date,
                authors=list({c.author_login, c.committer_login}),
                text=c.message,
                draft=False,
                path=f"{c.sha[6:]}.md",
            )


if __name__ == "__main__":
    blog = MkGitBlog("phil65", "mknodes", "blog/posts")
    blog.add_commits()
    page = blog.nav.pages[1]

    print(page)



from __future__ import annotations

import os

from mknodes.basenodes import mknode
from mknodes.utils import log, reprhelpers, resources


logger = log.get_logger(__name__)


class MkSnippet(mknode.MkNode):
    """Snippet to include markdown from another file.

    [More info](https://facelessuser.github.io/pymdown-extensions/extensions/snippets/)
    """

    REQUIRED_EXTENSIONS = [resources.Extension("pymdownx.snippets")]
    ICON = "material/paperclip"

    def __init__(self, path: str | os.PathLike, **kwargs):
        """Constructor.

        Arguments:
            path: Path to markdown file
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.path = path

    def __str__(self):
        return self.to_markdown()

    def __repr__(self):
        return reprhelpers.get_repr(self, path=self.path)

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        node = MkSnippet(path="README.md")
        page += mk.MkReprRawRendered(node)

    def _to_markdown(self) -> str:
        return f"--8<--\n{self.path}\n--8<--\n"


if __name__ == "__main__":
    section = MkSnippet("test.md", header="test")
    print(section.to_markdown())



from __future__ import annotations

import collections

from collections.abc import Mapping
import dataclasses
import functools
import importlib
from importlib import metadata
import types
from typing import Any

from packaging.markers import Marker
from packaging.requirements import Requirement
import pip._internal as pip

from mknodes.utils import log


logger = log.get_logger(__name__)


def install(package: str, editable: bool = False):
    """Pip-Install distribution with given options.

    Arguments:
        package: Name of the package to install
        editable: Whether to install in editable mode
    """
    cmd = ["install", "-e", package] if editable else ["install", package]
    pip.main(cmd)


def install_or_import(module_name: str) -> types.ModuleType:
    """If required, try to install given package and import it.

    This method relies on module name == distribution_name

    Arguments:
        module_name: Name of the module to import / install
    """
    try:
        return importlib.import_module(module_name)
    except ImportError:
        install(module_name)
        return importlib.import_module(module_name)


@functools.cache
def get_distribution(name: str) -> metadata.Distribution:
    """Cached version of metadata.distribution.

    Arguments:
        name: Name of the distribution to get an object for.
    """
    return metadata.distribution(name)


@functools.cache
def get_metadata(dist: metadata.Distribution):
    return dist.metadata


@functools.cache
def get_requires(dist: metadata.Distribution) -> list[str]:
    return dist.requires or []


@functools.cache
def get_package_map() -> Mapping[str, list[str]]:
    """Return a mapping of top-level packages to their distributions."""
    return metadata.packages_distributions()


@functools.cache
def distribution_to_package(dist: str):
    """Return the top-level package for given distribution.

    Arguments:
        dist: Name of the distribution to get the package for.
    """
    result = next((k for k, v in get_package_map().items() if dist in v), dist)
    return result.replace("-", "_").lower()


@functools.cache
def get_marker(marker):
    return Marker(marker)


@functools.cache
def _get_entry_points(dist: metadata.Distribution | None = None, **kwargs):
    return dist.entry_points if dist else metadata.entry_points(**kwargs)


def get_extras(markers: list) -> list[str]:
    extras = []
    for marker in markers:
        match marker:
            case list():
                extras.extend(get_extras(marker))
            case tuple():
                if str(marker[0]) == "extra":
                    extras.append(str(marker[2]))
    return extras


@dataclasses.dataclass
class EntryPoint:
    """EntryPoint including imported module."""

    name: str
    dotted_path: str
    group: str

    def load(self) -> Any:
        """Import and return the EntryPoint object."""
        if ":" in self.dotted_path:
            mod_name, kls_name = self.dotted_path.split(":")
        else:
            mod_name, kls_name = self.dotted_path, None
        mod = importlib.import_module(mod_name)
        return getattr(mod, kls_name) if kls_name else mod

    @property
    def module(self) -> str:
        """The module of the entry point."""
        return self.dotted_path.split(":")[0]

    @property
    def obj_path(self) -> str:
        """The dotted path of the object (without the module)."""
        return self.dotted_path.split(":")[1]


@functools.cache
def get_entry_points(
    dist: metadata.Distribution | str | None = None,
    group: str | None = None,
    **kwargs: Any,
) -> EntryPointMap:  # [str, list[EntryPoint]]
    """Returns a dictionary with entry point group as key, entry points as value.

    Arguments:
        dist: Optional distribution filter.
        group: Optional group filter.
        kwargs: Entry point filters
    """
    if dist:
        if isinstance(dist, str):
            dist = get_distribution(dist)
        eps = [i for i in _get_entry_points(dist) if i.group == group or not group]
    else:
        kw_args = dict(group=group, **kwargs) if group else kwargs
        eps = _get_entry_points(**kw_args)
    return EntryPointMap(eps)


class EntryPointMap(collections.defaultdict):
    def __init__(self, eps: list | None = None):
        super().__init__(list)
        for ep in eps or []:
            if not isinstance(ep, EntryPoint):
                ep = EntryPoint(name=ep.name, dotted_path=ep.value, group=ep.group)
            self[ep.group].append(ep)

    @classmethod
    def from_system(cls):
        eps = _get_entry_points()
        kls = cls()
        for group in eps.groups:
            kls[group] = [
                EntryPoint(name=ep.name, dotted_path=ep.value, group=ep.group)
                for ep in eps.select(group=group)
            ]
        # kls.update([i for ls in eps.select() for i in ls])
        return kls

    @property
    def all_eps(self):
        return [i for ls in self.values() for i in ls]

    def by_name(self, name: str) -> EntryPoint | None:
        return next((i for i in self.all_eps if i.name == name), None)

    def by_dotted_path(self, dotted_path: str) -> EntryPoint | None:
        return next((i for i in self.all_eps if i.dotted_path == dotted_path), None)


class Dependency:
    def __init__(self, name: str):
        self.req = Requirement(name)
        self.name = self.req.name
        self._marker = name

    @property
    def marker(self):
        return (
            get_marker(self._marker.split(";", maxsplit=1)[-1])
            if ";" in self._marker
            else None
        )

    @functools.cached_property
    def extras(self):
        return get_extras(self.marker._markers) if self.marker else []

    def __repr__(self):
        return f"{type(self).__name__}(name={self.name!r})"


@functools.cache
def get_dependency(name) -> Dependency:
    return Dependency(name)


@functools.cache
def list_installed_packages() -> dict[str, str]:
    import pkg_resources

    return {i.project_name: i.key for i in pkg_resources.working_set}


if __name__ == "__main__":
    eps = list_installed_packages()
    print(eps)



from __future__ import annotations

from collections.abc import Iterator
import contextlib

from typing import Any

from mknodes.basenodes import mknode
from mknodes.utils import log, reprhelpers


logger = log.get_logger(__name__)


class MkContainer(mknode.MkNode):
    """A base class for Nodes containing other MkNodes."""

    ICON = "material/database"

    def __init__(
        self,
        content: list | None | str | mknode.MkNode = None,
        block_separator: str | None = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            content: Child Nodes of this container
            block_separator: Separator to put between blocks. Defaults to 2 linebreaks
            kwargs: Keyword arguments passed to parent
        """
        super().__init__(**kwargs)
        self.block_separator = "\n\n" if block_separator is None else block_separator
        match content:
            case None:
                self.items: list[mknode.MkNode] = []
            case str():
                self.items = [self.to_child_node(content)] if content else []
            case mknode.MkNode():
                self.items = [self.to_child_node(content)]
            case list():
                self.items = [self.to_child_node(i) for i in content]
            case _:
                raise TypeError(content)

    def __bool__(self):
        return bool(self.items)

    def __add__(self, other: str | mknode.MkNode):
        self.append(other)
        return self

    def __iter__(self) -> Iterator[mknode.MkNode]:  # type: ignore
        return iter(self.items)

    def __repr__(self):
        content = [reprhelpers.to_str_if_textnode(i) for i in self.items]
        return reprhelpers.get_repr(self, content=content)

    @contextlib.contextmanager
    def in_html_tag(self, tag_name: str, **attributes: Any):
        """Will wrap the content in html tags with tag name.

        Examples:
            with node.in_html_tag("div", **{"class": "css_class"}):
                node += mk.MkSomething()

        Arguments:
            tag_name: Tag to use for wrapping
            attributes: addtional attributes for the XML element
        """
        if attributes:
            attr_txt = " " + " ".join(f"{k!r}={v!r}" for k, v in attributes.items())
        else:
            attr_txt = ""
        text = f"<{tag_name}{attr_txt}>"
        self.append(text)
        container = MkContainer(parent=self)
        yield container
        for item in container:
            self.append(item)
        self.append(f"</{tag_name}>")

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        page += "MkContainers are usually only used as a base class"
        page += "It basically only carries other nodes and stringifies them sequentially."
        item_1 = mk.MkCode(code="a = 1 + 2")
        item_2 = mk.MkText("abc")
        node = MkContainer(content=[item_1, item_2])
        page += mk.MkReprRawRendered(node)

    def _to_markdown(self) -> str:
        return self.block_separator.join(i.to_markdown() for i in self.items)

    def append(self, other: str | mknode.MkNode):
        """Append a MkNode to the end of given MkPage.

        Arguments:
            other: The node / text to append
        """
        node = self.to_child_node(other)
        self.items.append(node)  # type: ignore[arg-type]

    def insert(self, index: int, other: str | mknode.MkNode):
        """Insert a MkNode into desired position of given MkPage.

        Arguments:
            index: Position where node should get inserted
            other: The node / text to insert
        """
        node = self.to_child_node(other)
        self.items.insert(index, node)

    @property  # type: ignore
    def children(self) -> list[mknode.MkNode]:
        return self.items

    @children.setter
    def children(self, children: list[mknode.MkNode]):
        self.items = children


if __name__ == "__main__":
    section = MkContainer(header="fff")
    with section.in_html_tag("ab", c="e"):
        section += "hello"
    print(section)



from __future__ import annotations

from mknodes.basenodes import mktabcontainer, mktabs
from mknodes.utils import log, resources


logger = log.get_logger(__name__)


class MkTabbed(mktabcontainer.MkTabContainer):
    """PyMdown-based Tab container."""

    items: list[mktabs.MkTab]
    REQUIRED_EXTENSIONS = [
        resources.Extension("pymdownx.tabbed"),
        resources.Extension("pymdownx.superfences"),
    ]
    Tab = mktabs.MkTab

    @classmethod
    def create_example_page(cls, page):
        import mknodes as mk

        # this node is basically a container and manager for MkTabs nodes.
        node = MkTabbed(tabs={"Tab 1": "Some markdown", "Tab 2": "Other Markdown"})
        page += mk.MkReprRawRendered(node, header="### Regular")
        admonition = mk.MkAdmonition("Nested admonition")
        nested_node = MkTabbed(tabs={"Tabs": node, "Admonition": admonition})
        page += mk.MkReprRawRendered(nested_node, header="### Nested")


if __name__ == "__main__":
    tabs = dict(Tab1="Some text", Tab2="Another text")
    tabbed = MkTabbed(tabs)
    print(tabbed)



from __future__ import annotations

import mknodes as mk


EXPECTED = """/// tab | abc
    new: True

bcd
///
"""

NESTED_EXPECTED = """/// outer

//// inner_1

inner_1 content
////


//// inner_2

inner_2 content
////
///
"""


def test_block():
    node = mk.MkBlock(
        "tab",
        argument="abc",
        content="bcd",
        attributes=dict(new=True),
    )
    assert str(node) == EXPECTED


def test_nested_block():
    inner_1 = mk.MkBlock("inner_1", content="inner_1 content")
    inner_2 = mk.MkBlock("inner_2", content="inner_2 content")
    outer = mk.MkBlock("outer", content=[inner_1, inner_2])
    assert str(outer) == NESTED_EXPECTED



from __future__ import annotations

from collections.abc import Mapping
import contextlib
import os
import pathlib

from typing import Any

import jinja2

from mknodes.jinja import loaders, undefined as undefined_
from mknodes.utils import jinjahelpers, log, mergehelpers, pathhelpers


logger = log.get_logger(__name__)


class Environment(jinja2.Environment):
    """An enhanced Jinja environment."""

    def __init__(
        self,
        *,
        undefined: undefined_.UndefinedStr | type[jinja2.Undefined] = "strict",
        trim_blocks: bool = True,
        load_templates: bool = False,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            undefined: Handling of "Undefined" errors
            trim_blocks: Whitespace handling. Changes jinja default to `True`.
            load_templates: Whether to load the templates into environment.
            kwargs: Keyword arguments passed to parent
        """
        loader = loaders.resource_loader if load_templates else None
        if isinstance(undefined, str):
            undefined = undefined_.UNDEFINED_BEHAVIOR[undefined]
        self._extra_files: set[str] = set()
        self._extra_paths: set[str] = set()
        super().__init__(
            undefined=undefined,
            loader=loader,
            trim_blocks=trim_blocks,
            **kwargs,
        )
        self.filters.update(jinjahelpers.get_filters())
        self.globals.update(jinjahelpers.get_globals())
        self.filters["render_template"] = self.render_template
        self.filters["render_string"] = self.render_string
        self.filters["render_file"] = self.render_file

    def __contains__(self, template: str | os.PathLike):
        return pathlib.Path(template).as_posix() in self.list_templates()

    def __getitem__(self, val: str) -> jinja2.Template:
        return self.get_template(val)

    def inherit_from(self, env: jinja2.Environment):
        self.__dict__.update(env.__dict__)
        self.linked_to = env
        self.overlayed = True

    def merge_globals(self, other: Mapping, additive: bool = False):
        """Merge other into the environment globals with given strategy.

        Arguments:
            other: Globals to merge into environment
            additive: Whether an additive strategy should be used instead of replace.
        """
        strategy = "additive" if additive else "replace"
        mapping = mergehelpers.merge_dicts(self.variables, other, strategy=strategy)
        self.variables = dict(mapping)

    def add_template(self, file: str | os.PathLike):
        """Add a new template during runtime.

        Will create a new DictLoader and inject it into the existing loaders.

        Useful since render_string/render_file does not allow to use a parent template.
        Using this, render_template can be used.

        Arguments:
            file: File to add as a template
        """
        # we keep track of already added extra files to not add things multiple times.
        file = str(file)
        if file in self._extra_files:
            return
        self._extra_files.add(file)
        content = pathhelpers.load_file_cached(file)
        new_loader = loaders.DictLoader({file: content})
        self._add_loader(new_loader)

    def add_template_path(self, *path: str | os.PathLike):
        """Add a new template path during runtime.

        Will append a new FileSystemLoader by wrapping it and the the current loader into
        either an already-existing or a new Choiceloader.

        Arguments:
            path: Template serch path(s) to add
        """
        for p in path:
            if p in self._extra_paths:
                return
            self._extra_paths.add(str(p))
            new_loader = loaders.FileSystemLoader(p)
            self._add_loader(new_loader)

    def _add_loader(self, new_loader: jinja2.BaseLoader | dict | str | os.PathLike):
        match new_loader:
            case dict():
                new_loader = loaders.DictLoader(new_loader)
            case str() | os.PathLike():
                new_loader = loaders.FileSystemLoader(new_loader)
        match self.loader:
            case jinja2.ChoiceLoader():
                self.loader.loaders = [new_loader, *self.loader.loaders]
            case None:
                self.loader = new_loader
            case _:
                self.loader = loaders.ChoiceLoader(loaders=[new_loader, self.loader])

    def render_string(self, markdown: str, variables: dict | None = None):
        """Render a template string.

        Arguments:
            markdown: String to render
            variables: Extra variables for the environment
        """
        template = self.from_string(markdown)
        variables = variables or {}
        return template.render(**variables)

    def render_file(self, file: str | os.PathLike, variables: dict | None = None) -> str:
        """Helper to directly render a template from filesystem.

        Note: The file we pull in gets cached. That should be fine for our case though.

        Arguments:
            file: Template file to load
            variables: Extra variables for the environment
        """
        content = pathhelpers.load_file_cached(str(file))
        return self.render_string(content, variables)

    def render_template(
        self,
        template_name: str,
        variables: dict[str, Any] | None = None,
        block_name: str | None = None,
        parent_template: str | None = None,
    ) -> str:
        """Render a loaded template (or a block of a template).

        Arguments:
            template_name: Template name
            variables: Extra variables for this render call
            block_name: Render specific block from the template
            parent_template: Optional parent template (to be used with super())
        """
        template = self.get_template(template_name, parent=parent_template)
        if not block_name:
            variables = variables or {}
            return template.render(**variables)
        try:
            block_render_func = template.blocks[block_name]
        except KeyError:
            raise BlockNotFoundError(block_name, template_name) from KeyError

        ctx = template.new_context(variables or {})
        return self.concat(block_render_func(ctx))  # type: ignore
        # except Exception:
        #     self.handle_exception()

    @contextlib.contextmanager
    def with_globals(self, **kwargs: Any):
        """Context manager to temporarily set globals for the environment.

        Arguments:
            kwargs: Globals to set
        """
        temp = {}
        for k, v in kwargs.items():
            temp[k] = self.globals.get(k)
            self.globals[k] = v
        yield
        self.globals.update(temp)

    @contextlib.contextmanager
    def with_fence(
        self,
        start_variable: str = r"{{",
        end_variable: str = r"}}",
        start_block: str = "{%",
        end_block: str = "%}",
    ):
        """Context manager to temporarily set custom fences for jinja blocks.

        Arguments:
            start_variable: The string marking the beginning of a print statement
            end_variable: The string marking the end of a print statement
            start_block: The string marking the end of a block
            end_block: The string marking the end of a block
        """
        old_start_block = self.block_start_string
        old_end_block = self.block_end_string
        old_start_var = self.variable_start_string
        old_end_var = self.variable_end_string
        self.block_start_string = start_block
        self.block_end_string = end_block
        self.variable_start_string = start_variable
        self.variable_end_string = end_variable
        yield
        self.block_start_string = old_start_block
        self.block_end_string = old_end_block
        self.variable_start_string = old_start_var
        self.variable_end_string = old_end_var


class BlockNotFoundError(Exception):
    def __init__(
        self,
        block_name: str,
        template_name: str,
        message: str | None = None,
    ):
        self.block_name = block_name
        self.template_name = template_name
        super().__init__(
            message
            or f"Block {self.block_name!r} not found in template {self.template_name!r}",
        )


if __name__ == "__main__":
    env = Environment()
    txt = """{% filter styled(bold=True) %}
    test
    {% endfilter %}
    """
    print(env.render_string(txt))
    # text = env.render_string(r"{{ 'test' | MkHeader }}")
    # text = env.render_string(r"{{ 50 | MkProgressBar }}")
    # env.render_string(r"{{test('hallo')}}")
